from builtins import (range, super)

import numpy

from . import abstract
from . import mixin

class RaggedIndexedArray(mixin.RaggedIndexed,
                         abstract.CompressedArray):
    '''An underlying indexed ragged array.

A collection of features stored using an indexed ragged array combines
all features along a single dimension (the "sample dimension") such
that the values of each feature in the collection are interleaved.

The information needed to uncompress the data is stored in an "index
variable" that specifies the feature that each element of the sample
dimension belongs to.

.. versionadded:: 1.7.0

    '''
    def __init__(self, compressed_array=None, shape=None, size=None,
                 ndim=None, index_variable=None):
        '''**Initialization**

:Parameters:

    compressed_array: subclass of `Array`
        The compressed array.

    shape: `tuple`
        The uncompressed array dimension sizes.

    size: `int`
        Number of elements in the uncompressed array.

    ndim: `int`
        The number of uncompressed array dimensions

    index_variable: `Index`
        The index variable required to uncompress the data,
        corresponding to a CF-netCDF index variable.

        '''
        super().__init__(compressed_array=compressed_array,
                         shape=shape, size=size, ndim=ndim,
                         index_variable=index_variable,
                         compressed_dimension=0,
                         compression_type='ragged indexed')
    #--- End: def

    def __getitem__(self, indices):
        '''x.__getitem__(indices) <==> x[indices]

Returns an subspace of the uncompressed data an independent numpy
array.

The indices that define the subspace are relative to the uncompressed
data and must be either `Ellipsis` or a sequence that contains an
index for each dimension. In the latter case, each dimension's index
must either be a `slice` object or a sequence of two or more integers.

Indexing is similar to numpy indexing. The only difference to numpy
indexing (given the restrictions on the type of indices allowed) is:

  * When two or more dimension's indices are sequences of integers
    then these indices work independently along each dimension
    (similar to the way vector subscripts work in Fortran).

        '''
        # ------------------------------------------------------------
        # Method: Uncompress the entire array and then subspace it
        # ------------------------------------------------------------
        
        compressed_array = self._get_compressed_Array()

        # Initialise the un-sliced uncompressed array
        uarray = numpy.ma.masked_all(self.shape, dtype=self.dtype)

        # --------------------------------------------------------
        # Compression by indexed ragged array.
        #
        # The uncompressed array has dimensions (instance
        # dimension, element dimension).
        # --------------------------------------------------------
        index_array = self.get_index_variable().get_array()
        
        for i in range(uarray.shape[0]):
            sample_dimension_indices = numpy.where(index_array == i)[0]

            u_indices = (i, #slice(i, i+1),
                         slice(0, len(sample_dimension_indices)))
            
            uarray[u_indices] = compressed_array[(sample_dimension_indices,)]
        #--- End: for

        return self.get_subspace(uarray, indices, copy=True)
    #--- End: def

#--- End: class



from builtins import super

from copy import deepcopy

from . import mixin
from . import core


class Bounds(mixin.NetCDFVariable,
             mixin.PropertiesData,
             core.Bounds):
    '''A cell bounds component of a coordinate or domain ancillary
construct of the CF data model.

An array of cell bounds spans the same domain axes as its coordinate
array, with the addition of an extra dimension whose size is that of
the number of vertices of each cell. This extra dimension does not
correspond to a domain axis construct since it does not relate to an
independent axis of the domain. Note that, for climatological time
axes, the bounds are interpreted in a special way indicated by the
cell method constructs.

In the CF data model, a bounds component does not have its own
properties because they can not logically be different to those of the
coordinate construct itself. However, it is sometimes desired to store
attributes on a CF-netCDF bounds variable, so it is also allowed to
provide properties to a bounds component.

.. versionadded:: 1.7.0

    '''
    def __init__(self, properties=None, data=None, source=None,
                 copy=True, _use_data=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

          *Parameter example:*
             ``properties={'standard_name': 'grid_latitude'}``
        
        Properties may also be set after initialisation with the
        `properties` and `set_property` methods.

    data: `Data`, optional
        Set the data. Ignored if the *source* parameter is set.
        
        The data also may be set after initialisation with the
        `set_data` method.
        
    source: optional
        Initialize the properties and data from those of *source*.
        
    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(properties=properties, data=data,
                         source=source, copy=copy,
                         _use_data=_use_data)
        
        if source is not None:
            try:
                inherited_properties = source.inherited_properties()
            except AttributeError:
                inherited_properties = {}
        else:
            inherited_properties = {}

        self._set_component('inherited_properties', inherited_properties)

        self._initialise_netcdf(source)
    #--- End: def
    
    def dump(self, display=True, field=None, key=None, _title=None,
             _create_title=True, _prefix=None, _level=0,
             _omit_properties=None, _axes=None, _axis_names=None):
        '''A full description of the bounds component.

Returns a description of all properties and provides selected values
of all data arrays.

.. versionadded:: 1.7.0

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.

:Returns:

    `None` or `str`
        The description. If *display* is True then the description is
        printed and `None` is returned. Otherwise the description is
        returned as a string.

        '''
        if _create_title and _title is None: 
            _title = 'Bounds: ' + self.name(default='')

        return super().dump(display=display, field=field, key=key,
                            _omit_properties=_omit_properties,
                            _prefix=_prefix, _level=_level,
                            _title=_title,
                            _create_title=_create_title, _axes=_axes,
                            _axis_names=_axis_names)
    #--- End: def
    
    def get_data(self, *default):
        '''Return the data.

Note that the data are returned in a `Data` object. Use the
`get_array` method to return the data as an independent `numpy` array.

.. versionadded:: 1.7.0

.. seealso:: `data`, `del_data`, `get_array`, `has_data`, `set_data`

:Parameters:

    default: optional
        Return *default* if a data object has not been set.

:Returns:

        The data object. If unset then *default* is returned, if
        provided.

**Examples:**

>>> d = cfdm.Data(range(10))
>>> f.set_data(d)
>>> f.has_data()
True
>>> f.get_data()
<Data(10): [0, ..., 9]>
>>> f.del_data()
<Data(10): [0, ..., 9]>
>>> f.has_data()
False
>>> print(f.get_data(None))
None
>>> print(f.del_data(None))
None

        '''
        data = super().get_data(None)

        if data is None:
            return super().get_data(*default)

        if data.get_units(None) is None:
            units = self.inherited_properties().get('units')
            if units is not None:
                data.set_units(units)
        #--- End: if
    
        if data.get_calendar(None) is None:
            calendar = self.inherited_properties().get('calendar')
            if calendar is not None:
                data.set_calendar(calendar)
        #--- End: if

        return data
    #--- End: def

    def inherited_properties(self):
        '''Return the properties inherited from a coordinate construct.

.. versionadded:: 1.7.0

.. seealso:: `properties`

:Returns:

    `dict`
        The inherited properties.

**Examples:**

>>> b.properties()
{}
>>> b.inherited_properties()
{'standard_name': 'longitude',
 'units': 'degrees_east'}

        '''
        return deepcopy(self._get_component('inherited_properties', {}))
    #--- End: def   

    def name(self, default=None, ncvar=True, custom=None,
             all_names=False):
        '''Return a name.

By default the name is the first found of the following:

1. The "standard_name" property.
2. The "cf_role" property, preceeded by ``'cf_role:'``.
3. The "long_name" property, preceeded by ``'long_name:'``.
4. The netCDF variable name, preceeded by ``'ncvar%'``.
5. The value of the *default* parameter.

.. versionadded:: 1.7.0

:Parameters:

    default: optional
        If no other name can be found then return the value of the
        *default* parameter. By default `None` is returned in this
        case.

    ncvar: `bool`, optional
        If False then do not consider the netCDF variable name.

    all_names: `bool`, optional
        If True then return a list of all possible names.

    custom: sequence of `str`, optional
        Replace the ordered list of properties from which to seatch
        for a name. The default list is ``['standard_name', 'cf_role',
        'long_name']``.

        *Parameter example:*
          ``custom=['project']``

        *Parameter example:*
          ``custom=['project', 'long_name']``

:Returns:

        The name. If the *all_names* parameters is True then a list of
        all possible names.

**Examples:**

>>> f.properties()
{'foo': 'bar',
 'long_name': 'Air Temperature',
 'standard_name': 'air_temperature'}
>>> f.nc_get_variable()
'tas'
>>> f.name()
'air_temperature'
>>> f.name(all_names=True)
['air_temperature', 'long_name:Air Temperature', 'ncvar:tas']
>>> x = f.del_property('standard_name')
>>> f.name()
'long_name:Air Temperature'
>>> x = f.del_property('long_name')
>>> f.name()
'ncvar:tas'
>>> f.name(custom=['foo'])
'foo:bar'
>>> f.name(default='no name', custom=['foo'])
['foo:bar', 'no name']

        '''
        inherited_properties = self.inherited_properties()
        if inherited_properties:
            properties = inherited_properties.copy()
            properties.update(self.properties())

            bounds = self.copy()
            bounds.properties(properties)
            self = bounds
            
        return super().name(default=default, ncvar=ncvar,
                            custom=custom, all_names=all_names)
    #--- End: def

#--- End: class



from __future__ import print_function
from builtins import (str, super)

from . import mixin
from . import core

from . import Constructs


class Domain(mixin.ConstructAccess,
             mixin.Container,
             core.Domain):
    '''A domain of the CF data model.

The domain represents a set of discrete "locations" in what generally
would be a multi-dimensional space, either in the real world or in a
model's simulated world. These locations correspond to individual data
array elements of a field construct

The domain is defined collectively by the following constructs of the
CF data model: domain axis, dimension coordinate, auxiliary
coordinate, cell measure, coordinate reference and domain ancillary
constructs.

.. versionadded:: 1.7.0

    '''
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._Constructs = Constructs
        return instance
    #--- End: def
    
    def __repr__(self):
        '''Called by the `repr` built-in function.

x.__repr__() <==> repr(x)

        '''
        shape = sorted([domain_axis.get_size() for domain_axis in list(self.domain_axes().values())])
        shape = str(shape)
        shape = shape[1:-1]
        
        return '<{0}: {{{1}}}>'.format(self.__class__.__name__, shape)
    #--- End: def

    def __str__(self):
        '''Called by the `str` built-in function.

x.__str__() <==> str(x)

        '''
        def _print_item(self, cid, variable, axes, dimension_coord):
            '''Private function called by __str__'''
            
            if dimension_coord:
                # Dimension coordinate
                name = variable.name(ncvar=True, default=cid)
                if variable.has_data():
                    name += '({0})'.format(variable.get_data().size)
                elif hasattr(variable, 'nc_external'):
                    if variable.nc_external():
                        ncvar = variable.nc_get_variable(None)
                        if ncvar is not None:
                            x.append(' (external variable: ncvar%{})'.format(ncvar))
                        else:
                            x.append(' (external variable)')
                            
                if variable is None:
                    return name
                          
                x = [name]
                
            else:
                # Auxiliary coordinate
                # Cell measure
                # Field ancillary
                # Domain ancillary
                x = [variable.name(ncvar=True, default=cid)]

                if variable.has_data():
#                    shape = [axis_names[axis] for axis in self.constructs_data_axes(cid)]
                    shape = [axis_names[axis] for axis in axes]
                    shape = str(tuple(shape)).replace("'", "")
                    shape = shape.replace(',)', ')')
                    x.append(shape)
                elif hasattr(variable, 'nc_external'):
                    if variable.nc_external():
                        ncvar = variable.nc_get_variable(None)
                        if ncvar is not None:
                            x.append(' (external variable: ncvar%{})'.format(ncvar))
                        else:
                            x.append(' (external variable)')
            #--- End: if
                
            if variable.has_data():
                x.append(' = {0}'.format(variable.get_data()))
                
            return ''.join(x)
        #--- End: def
                          
        string = []
        
        axis_name = self.domain_axis_name

        axis_names = self._unique_domain_axis_names()

        constructs_data_axes = self.constructs_data_axes()
        
        x = []
        for axis_cid in sorted(self.domain_axes()):
            for cid, dim in list(self.dimension_coordinates().items()):
#                if self.constructs_data_axes()[cid] == (axis_cid,):
                if constructs_data_axes[cid] == (axis_cid,):
                    name = dim.name(default='cid%{0}'.format(cid), ncvar=True)
                    y = '{0}({1})'.format(name, dim.get_data().size)
                    if y != axis_names[axis_cid]:
                        y = '{0}({1})'.format(name, axis_names[axis_cid])
                    if dim.has_data():
                        y += ' = {0}'.format(dim.get_data())
                        
                    x.append(y)
        #--- End: for
        if x:
            string.append('Dimension coords: {}'.format('\n                : '.join(x)))

        # Auxiliary coordinates
        x = [_print_item(self, cid, v, constructs_data_axes[cid], False) 
             for cid, v in sorted(self.auxiliary_coordinates().items())]
        if x:
            string.append('Auxiliary coords: {}'.format(
                '\n                : '.join(x)))
        
        # Cell measures
        x = [_print_item(self, cid, v, constructs_data_axes[cid], False)
             for cid, v in sorted(self.cell_measures().items())]
        if x:
            string.append('Cell measures   : {}'.format(
                '\n                : '.join(x)))
            
        # Coordinate references
        x = sorted([str(ref) for ref in list(self.coordinate_references().values())])
        if x:
            string.append('Coord references: {}'.format(
                '\n                : '.join(x)))
            
        # Domain ancillary variables
        x = [_print_item(self, cid, anc, constructs_data_axes[cid], False)
             for cid, anc in sorted(self.domain_ancillaries().items())]
        if x:
            string.append('Domain ancils   : {}'.format(
                '\n                : '.join(x)))
                                      
        string.append('')
        
        return '\n'.join(string)
    #--- End def
                          
    def _dump_axes(self, axis_names, display=True, _level=0):
        '''Return a string containing a description of the domain axes of the
field.
    
:Parameters:
    
    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.
    
    _level: `int`, optional

:Returns:
    
    `str`
        A string containing the description.
    
**Examples:**

        '''
        indent1 = '    ' * _level
        indent2 = '    ' * (_level+1)

        axes = self.domain_axes()

        w = sorted(["{0}Domain Axis: {1}".format(indent1, axis_names[axis])
                    for axis in axes])

        string = '\n'.join(w)

        if display:
            print(string)
        else:
            return string
    #--- End: def

    def del_construct(self, description=None, cid=None, axes=None,
                      construct_type=None, default=ValueError()):
        '''Remove a metadata construct.

The construct is identified via optional parameters. The *unique*
construct that satisfies *all* of the given criteria is removed.

If a domain axis construct is to be removed then it can't be spanned
by any data arrays.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `get_construct`, `has_construct`,
             `set_construct`

:Parameters:

    description: `str`, optional
        Select constructs that have the given property, or other
        attribute, value.

        The description may be one of:

        * The value of the standard name property on its own. 

          *Parameter example:*
            ``description='air_pressure'`` will select constructs that
            have a "standard_name" property with the value
            "air_pressure".

        * The value of any property prefixed by the property name and
          a colon (``:``).

          *Parameter example:*
            ``description='positive:up'`` will select constructs that
            have a "positive" property with the value "up".

          *Parameter example:*
            ``description='foo:bar'`` will select constructs that have
            a "foo" property with the value "bar".

          *Parameter example:*
            ``description='standard_name:air_pressure'`` will select
            constructs that have a "standard_name" property with the
            value "air_pressure".

        * The measure of cell measure constructs, prefixed by
          ``measure%``.

          *Parameter example:*
            ``description='measure%area'`` will select "area" cell
            measure constructs.

        * A construct identifier, prefixed by ``cid%`` (see also the
          *cid* parameter).

          *Parameter example:* 
            ``description='cid%domainancillary1'`` will select domain
            ancillary construct with construct identifier
            "domainancillary1". This is equivalent to
            ``cid='domainancillary1'``.

        * The netCDF variable name, prefixed by ``ncvar%``.

          *Parameter example:*
            ``description='ncvar%lat'`` will select constructs with
              netCDF variable name "lat".

        * The netCDF dimension name of domain axis constructs,
          prefixed by ``ncdim%``.

          *Parameter example:*
            ``description='ncdim%time'`` will select domain axis
            constructs with netCDF dimension name "time".

    cid: `str`, optional
        Select the construct with the given construct identifier.

        *Parameter example:*
          ``cid='domainancillary0'`` will the domain ancillary
          construct with construct identifier "domainancillary1". This
          is equivalent to ``description='cid%domainancillary0'``.

    construct_type: `str`, optional
        Select constructs of the given type. Valid types are:

          ==========================  ================================
          *construct_type*            Constructs
          ==========================  ================================
          ``'domain_ancillary'``      Domain ancillary constructs
          ``'dimension_coordinate'``  Dimension coordinate constructs
          ``'domain_axis'``           Domain axis constructs
          ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
          ``'cell_measure'``          Cell measure constructs
          ``'coordinate_reference'``  Coordinate reference constructs
          ==========================  ================================

        *Parameter example:*
          ``construct_type='dimension_coordinate'``

        *Parameter example:*
          ``construct_type=['auxiliary_coordinate']``

        *Parameter example:*
          ``construct_type=('domain_ancillary', 'cell_measure')``

    axes: sequence of `str`, optional
        Select constructs which have data that spans one or more of
        the given domain axes, in any order. Domain axes are specified
        by their construct identifiers.

        *Parameter example:*
          ``axes=['domainaxis2']``

        *Parameter example:*
          ``axes=['domainaxis0', 'domainaxis1']``

    default: optional
        Return *default* if no metadata construct can be found. By
        default an exception is raised in this case.

        *Parameter example:*
          ``default=None``

:Returns:

        The removed metadata construct.

**Examples:**

>>> c = f.del_construct('grid_latitude')
>>> c = f.del_construct('long_name:Air Pressure')
>>> c = f.del_construct('ncvar%lat)
>>> c = f.del_construct('cid%cellmeasure0')
>>> c = f.del_construct(cid='domainaxis2')
>>> c = f.del_construct(construct_type='auxiliary_coordinate',
...                     axes=['domainaxis1'])

        '''
        cid = self.get_construct_id(description=description, cid=cid,
                                    construct_type=construct_type,
                                    axes=axes, default=None)
        if cid is None:
            return self._get_constructs()._default(
                default, 'No unique construct meets criteria')
            
        return self._get_constructs().del_construct(cid=cid)
    #--- End: def

    def domain_axis_name(self, cid):
        '''TODO
        '''
        constructs = self._get_constructs()
        return constructs.domain_axis_name(cid)
    #--- End: def

    def dump(self, display=True, _level=0, _title=None):
        '''A full description of the domain.

The domain components are described without abbreviation with the
exception of data arrays, which are abbreviated to their first and
last values.

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.

          *Parameter example:*
            ``f.dump()`` is equivalent to ``print
            f.dump(display=False)``.

:Returns:

    `None` or `str`
        If *display* is True then the description is printed and
        `None` is returned. Otherwise the description is returned as a
        string.

        '''
        indent = '    '      
        indent0 = indent * _level
        indent1 = indent0 + indent

        axis_to_name = self._unique_domain_axis_names()

        construct_name = self._unique_construct_names()

        constructs_data_axes = self.constructs_data_axes()
        
        string = []

        # Domain axes
        axes = self._dump_axes(axis_to_name, display=False, _level=_level)
        if axes:
            string.append(axes)
          
        # Dimension coordinates
        for cid, value in sorted(self.dimension_coordinates().items()):
            string.append('')
            string.append(
                value.dump(display=False, _level=_level,
                           _title='Dimension coordinate: {0}'.format(
                               construct_name[cid]),
                           _axes=constructs_data_axes[cid],
                           _axis_names=axis_to_name))
            
        # Auxiliary coordinates
        for cid, value in sorted(self.auxiliary_coordinates().items()):
            string.append('')
            string.append(
                value.dump(display=False, _level=_level,
                           _title='Auxiliary coordinate: {0}'.format(
                               construct_name[cid]),
                           _axes=constructs_data_axes[cid],
                           _axis_names=axis_to_name))

        # Domain ancillaries
        for cid, value in sorted(self.domain_ancillaries().items()):
            string.append('') 
            string.append(value.dump(display=False, _level=_level,
                                     _title='Domain ancillary: {0}'.format(
                                         construct_name[cid]),
                                     _axes=constructs_data_axes[cid],
                                     _axis_names=axis_to_name))
            
        # Coordinate references
        for cid, value in sorted(self.coordinate_references().items()):
            string.append('')
            string.append(
                value.dump(display=False, _level=_level,
                           _title='Coordinate reference: {0}'.format(
                               construct_name[cid]),
                           _construct_names=construct_name,
                           _auxiliary_coordinates=tuple(self.auxiliary_coordinates()),
                           _dimension_coordinates=tuple(self.dimension_coordinates())))
            
        # Cell measures
        for cid, value in sorted(self.cell_measures().items()):
            string.append('')
            string.append(
                value.dump(display=False, field=self,
                           key=cid, _level=_level,
                           _title='Cell measure: {0}'.format(construct_name[cid]),
                           _axes=constructs_data_axes[cid],
                           _axis_names=axis_to_name))
            

        string.append('')
        
        string = '\n'.join(string)
       
        if display:
            print(string)
        else:
            return string
    #--- End: def

    def equals(self, other, rtol=None, atol=None, verbose=False,
               ignore_data_type=False, ignore_fill_value=False,
               ignore_properties=(), ignore_type=False):
        '''TODO

.. versionadded:: 1.7.0
        '''
        ignore_properties = tuple(ignore_properties) + ('Conventions',)
            
        if not super().equals(other, rtol=rtol, atol=atol,
                              verbose=verbose,
                              ignore_data_type=ignore_data_type,
                              ignore_fill_value=ignore_fill_value,
                              ignore_properties=ignore_properties,
                              ignore_type=ignore_type):
            return False

        # ------------------------------------------------------------
        # Check the constructs
        # ------------------------------------------------------------              
        if not self._equals(self._get_constructs(), other._get_constructs(),
                            rtol=rtol, atol=atol,
                            verbose=verbose,
                            ignore_data_type=ignore_data_type,
                            ignore_type=ignore_type,
                            ignore_fill_value=ignore_fill_value):
            if verbose:
                print(
                    "{0}: Different {1}".format(self.__class__.__name__, 'constructs'))
            return False

        return True
    #--- End: def
        
#--- End: class



from . import PropertiesDataBounds


class Coordinate(PropertiesDataBounds):
    '''Mixin class for dimension or auxiliary coordinate constructs.

.. versionadded:: 1.7.0

    '''
    
#--- End: class



from builtins import super

from . import abstract


class Bounds(abstract.PropertiesData):
    '''A cell bounds component of a coordinate or domain ancillary
construct of the CF data model.

An array of cell bounds spans the same domain axes as its coordinate
array, with the addition of an extra dimension whose size is that of
the number of vertices of each cell. This extra dimension does not
correspond to a domain axis construct since it does not relate to an
independent axis of the domain. Note that, for climatological time
axes, the bounds are interpreted in a special way indicated by the
cell method constructs.

.. versionadded:: 1.7.0

    '''
    def __init__(self, properties=None, data=None, source=None,
                 copy=True, _use_data=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
           ``properties={'standard_name': 'altitude'}``
        
        Properties may also be set after initialisation with the
        `properties` and `set_property` methods.

    data: `Data`, optional
        Set the data. Ignored if the *source* parameter is set.
        
        The data also may be set after initialisation with the
        `set_data` method.
        
    source: optional
        Override the *properties* and *data* parameters with
        ``source.properties()`` and ``source.get_data(None)``
        respectively.

        If *source* does not have one of these methods, then the
        corresponding parameter is not set.
        
    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(properties=properties, data=data,
                         source=source, copy=copy)
    #--- End: def

#--- End: class



from distutils.version import LooseVersion

_CODE0 = {
        # Physically meaningful and corresponding to constructs
        'Cell measures variable' : 100,
        'cell_measures attribute': 101,
        
        'Bounds variable'        : 200,
        'bounds attribute'       : 201,
        
        'Ancillary variable': 120,
        'ancillary_variables attribute': 121,
        
        'Formula terms variable': 130,
        'formula_terms attribute': 131,
        'Bounds formula terms variable': 132,
        'Bounds formula_terms attribute': 133,
        
        'Auxiliary/scalar coordinate variable': 140,
        'coordinates attribute': 141,
        
        'grid mapping variable': 150,
        'grid_mapping attribute' : 151,
        'Grid mapping coordinate variable': 152, 
        
        'Cell method interval': 160,
        
        # Purely structural
        'Compressed dimension': 300,
        'compress attribute': 301,
        'Instance dimension':310,
        'instance_dimension attribute':311,
        'Count dimension': 320,
        'count_dimension attribute': 321,
    }
    
_CODE1 = {
        'is incorrectly formatted': 2,
        'is not in file': 3,
        'spans incorrect dimensions': 4,
        'is not in file nor referenced by the external_variables global attribute': 5,
        'has incompatible terms': 6,
        'that spans the vertical dimension has no bounds': 7,
    'that does not span the vertical dimension is inconsistent with the formula_terms of the parametric coordinate variable': 8,
    }

MAGIC_NUMBER = (21382211,
                1128547841,
                1178880137,
                38159427)

datum_parameters = ('earth_radius',
                    'geographic_crs_name',
                    'geoid_name',
                    'geopotential_datum_name',
                    'horizontal_datum_name',
                    'inverse_flattening',
                    'longitude_of_prime_meridian',
                    'prime_meridian_name',
                    'reference_ellipsoid_name',
                    'semi_major_axis',
                    'semi_minor_axis',
                    'towgs84',
)

# Col 0: Georeference name Col 1+: Set of coordinates to which it applies
coordinate_reference_coordinates = {
    'albers_conical_equal_area'                  : ('projection_x_coordinate',
                                                    'projection_y_coordinate',
                                                    'latitude',
                                                    'longitude',),
    'azimuthal_equidistant'                      : ('projection_x_coordinate',
                                                    'projection_y_coordinate',
                                                    'latitude',
                                                    'longitude',),
    'geostationary'                              : ('projection_x_coordinate',
                                                    'projection_y_coordinate',
                                                    'latitude',
                                                    'longitude',),
    'lambert_azimuthal_equal_area'               : ('projection_x_coordinate',
                                                    'projection_y_coordinate',
                                                    'latitude',
                                                    'longitude',),
    'lambert_conformal_conic'                    : ('projection_x_coordinate',
                                                    'projection_y_coordinate',
                                                    'latitude',
                                                    'longitude',),
    'lambert_cylindrical_equal_area'             : ('projection_x_coordinate',
                                                    'projection_y_coordinate',
                                                    'latitude',
                                                    'longitude',),
    'latitude_longitude'                         : ('latitude',
                                                    'longitude',),
    'mercator'                                   : ('projection_x_coordinate',
                                                    'projection_y_coordinate',
                                                    'latitude',
                                                    'longitude',),
    'orthographic'                               : ('projection_x_coordinate',
                                                    'projection_y_coordinate',
                                                    'latitude',
                                                    'longitude',),
    'polar_stereographic'                        : ('projection_x_coordinate',
                                                    'projection_y_coordinate',
                                                    'latitude',
                                                    'longitude',),
    'rotated_latitude_longitude'                 : ('grid_latitude',
                                                    'grid_longitude',
                                                    'latitude',
                                                    'longitude',),
    'sinusoidal'                                 : ('projection_x_coordinate',
                                                    'projection_y_coordinate',
                                                    'latitude',
                                                    'longitude',),
    'stereographic'                              : ('projection_x_coordinate',
                                                    'projection_y_coordinate',
                                                    'latitude',
                                                    'longitude',),
    'transverse_mercator'                        : ('projection_x_coordinate',
                                                    'projection_y_coordinate',
                                                    'latitude',
                                                    'longitude',),
    'vertical_perspective'                       : ('projection_x_coordinate',
                                                    'projection_y_coordinate',
                                                    'latitude',
                                                    'longitude',),
    'atmosphere_ln_pressure_coordinate'          : ('atmosphere_ln_pressure_coordinate',),
    'atmosphere_sigma_coordinate'                : ('atmosphere_sigma_coordinate',),
    'atmosphere_hybrid_sigma_pressure_coordinate': ('atmosphere_hybrid_sigma_pressure_coordinate',),
    'atmosphere_hybrid_height_coordinate'        : ('atmosphere_hybrid_height_coordinate',),
    'atmosphere_sleve_coordinate'                : ('atmosphere_sleve_coordinate',),
    'ocean_sigma_coordinate'                     : ('ocean_sigma_coordinate',),
    'ocean_s_coordinate'                         : ('ocean_s_coordinate',),
    'ocean_sigma_z_coordinate'                   : ('ocean_sigma_z_coordinate',),
    'ocean_double_sigma_coordinate'              : ('ocean_double_sigma_coordinate',),
}

description_of_file_contents_attributes = ('comment',
                                           'Conventions',
                                           'featureType',
                                           'history',
                                           'institution',
                                           'references',
                                           'source',
                                           'title',)

# --------------------------------------------------------------------
# Valid cell method properties
# --------------------------------------------------------------------
cell_method_properties = set(('within',
                              'where',
                              'over',
                              'interval',
                              'comment',
))



from builtins import object
from future.utils import with_metaclass
import abc


class Implementation(with_metaclass(abc.ABCMeta, object)):
    '''Store an implementation of the CF data model.
    '''
    
    def __init__(self, cf_version=None, **kwargs):
        '''**Initialisation**

:Parameters:

    kwargs:
        The concrete objects required to represent a Field.

        '''
        self._cf_version = cf_version
        self._class = kwargs.copy()
        for key, value in kwargs.items():
            if value is None:
                del self._class[key]
    #--- End: def

    def copy(self):
        '''Copy 
        
        '''
        return type(self)(self._class)
    #--- End: def

    def get_class(self, name):
        '''Return a class of the implementation.

:Parameters:

    name: `str`
        The name of the class.

          *Example:*
            ``name='Field'``

:Returns:
    
    out:
        The class object.

**Examples:**

>>> Field = i.get_class('Field')
>>> f = Field()

        '''
        try:
            return self._class[name]
        except KeyError:
            raise ValueError("Implementation does not have class {!r}".format(name))
    #--- End: def

    def get_cf_version(self):
        '''Return the CF version of the implementation.

:Returns:
    
    out: `str`
        The version.

**Examples:**

>>> i.get_cf_version()
'1.8'    

        '''
        return self._cf_version
    #--- End: def

    def set_class(self, name, cls):
        '''Set a class of the implementation.

:Parameters:

    name: `str`
        The name of the class.

          *Example:*
            ``name='Field'``
    cls: 
        The class object.

:Returns:

    `None`

        '''
        self._class[name] = cls
    #--- End: def

#--- End: class



from __future__ import print_function
import datetime
import os
import unittest

import numpy

import cfdm

class CoordinateReferenceTest(unittest.TestCase):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'test_file.nc')

    def test_CoordinateReference_equals(self):
        f = cfdm.read(self.filename)[0]

        # Create a vertical grid mapping coordinate reference
        t = cfdm.CoordinateReference(
            coordinates=('coord1',),
            coordinate_conversion=cfdm.CoordinateConversion(
                parameters={'standard_name': 'atmosphere_hybrid_height_coordinate'},
                domain_ancillaries={'a': 'aux0', 'b': 'aux1', 'orog': 'orog'})
        )
        self.assertTrue(t.equals(t.copy(), verbose=True))
        
        # Create a horizontal grid mapping coordinate reference
        t = cfdm.CoordinateReference(
            coordinates=['coord1', 'fred', 'coord3'],
             coordinate_conversion=cfdm.CoordinateConversion(
                 parameters={'grid_mapping_name': 'rotated_latitude_longitude',
                             'grid_north_pole_latitude': 38.0,
                             'grid_north_pole_longitude': 190.0})
        )            
        self.assertTrue(t.equals(t.copy(), verbose=True))

        datum=cfdm.Datum(parameters={'earth_radius': 6371007})
        conversion=cfdm.CoordinateConversion(
            parameters={'grid_mapping_name': 'rotated_latitude_longitude',
                        'grid_north_pole_latitude': 38.0,
                        'grid_north_pole_longitude': 190.0})
        
        t = cfdm.CoordinateReference(
            coordinate_conversion=conversion,
            datum=datum,
            coordinates=['x', 'y', 'lat', 'lon']
        )

        self.assertTrue(t.equals(t.copy(), verbose=True))

        # Create a horizontal grid mapping coordinate reference
        t = cfdm.CoordinateReference(
               coordinates=['coord1', 'fred', 'coord3'],
               coordinate_conversion=cfdm.CoordinateConversion(
                   parameters={'grid_mapping_name': 'albers_conical_equal_area',
                               'standard_parallel': [-30, 10],
                               'longitude_of_projection_origin': 34.8,
                               'false_easting': -20000,
                               'false_northing': -30000})
        )
        self.assertTrue(t.equals(t.copy(), verbose=True))

        # Create a horizontal grid mapping coordinate reference
        t = cfdm.CoordinateReference(
               coordinates=['coord1', 'fred', 'coord3'],
               coordinate_conversion=cfdm.CoordinateConversion(
                   parameters={'grid_mapping_name': 'albers_conical_equal_area',
                               'standard_parallel': cfdm.Data([-30, 10]),
                               'longitude_of_projection_origin': 34.8,
                               'false_easting': -20000,
                               'false_northing': -30000})
        )
        self.assertTrue(t.equals(t.copy(), verbose=True))
    #--- End: def

#--- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    print(cfdm.environment())
    print('')
    unittest.main(verbosity=2)



from __future__ import print_function
import datetime
import os
import tempfile
import time 
import unittest

import netCDF4
import numpy

import cfdm

def _make_files():
    '''
    '''
    def _pp(filename, parent=False, external=False, combined=False, external_missing=False):
        '''
        '''
        nc = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
        
        nc.createDimension('grid_latitude', 10)
        nc.createDimension('grid_longitude', 9)
        
        nc.Conventions = 'CF-1.7'
        if parent:
            nc.external_variables = 'areacella'

        if parent or combined or external_missing:
            grid_latitude = nc.createVariable(dimensions=('grid_latitude',),
                                              datatype='f8',
                                              varname='grid_latitude')
            grid_latitude.setncatts({'units': 'degrees', 'standard_name': 'grid_latitude'})
            grid_latitude[...] = range(10)
            
            grid_longitude = nc.createVariable(dimensions=('grid_longitude',),
                                               datatype='f8',
                                               varname='grid_longitude')
            grid_longitude.setncatts({'units': 'degrees', 'standard_name': 'grid_longitude'})
            grid_longitude[...] = range(9)
            
            latitude = nc.createVariable(dimensions=('grid_latitude', 'grid_longitude'),
                                         datatype='i4',
                                         varname='latitude')
            latitude.setncatts({'units': 'degree_N', 'standard_name': 'latitude'})
            
            latitude[...] = numpy.arange(90).reshape(10, 9)
            
            longitude = nc.createVariable(dimensions=('grid_longitude', 'grid_latitude'),
                                          datatype='i4',
                                          varname='longitude')
            longitude.setncatts({'units': 'degreeE', 'standard_name': 'longitude'})
            longitude[...] = numpy.arange(90).reshape(9, 10)
            
            eastward_wind = nc.createVariable(dimensions=('grid_latitude', 'grid_longitude'),
                                              datatype='f8',
                                              varname=u'eastward_wind')
            eastward_wind.coordinates = u'latitude longitude'
            eastward_wind.standard_name = 'eastward_wind'
            eastward_wind.cell_methods = 'grid_longitude: mean (interval: 1 day comment: ok) grid_latitude: maximum where sea'
            eastward_wind.cell_measures = 'area: areacella'
            eastward_wind.units = 'm s-1'
            eastward_wind[...] = numpy.arange(90).reshape(10, 9) - 45.5

        if external or combined:                
            areacella = nc.createVariable(dimensions=('grid_longitude', 'grid_latitude'),
                                          datatype='f8',
                                          varname='areacella')
            areacella.setncatts({'units': 'm2', 'standard_name': 'cell_area'})
            areacella[...] = numpy.arange(90).reshape(9, 10) + 100000.5
            
        nc.close()
    #--- End: def

    parent_file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                               'parent.nc')        
    external_file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                 'external.nc')            
    combined_file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                 'combined.nc')        
    external_missing_file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                         'external_missing.nc')            
    
    _pp(parent_file          , parent=True)
    _pp(external_file        , external=True)
    _pp(combined_file        , combined=True)
    _pp(external_missing_file, external_missing=True)

    return parent_file, external_file, combined_file, external_missing_file
#--- End: def

(parent_file,
 external_file,
 combined_file,
 external_missing_file) = _make_files()


class ExternalVariableTest(unittest.TestCase):
    def setUp(self):
        self.parent_file           = parent_file
        self.external_file         = external_file
        self.combined_file         = combined_file
        self.external_missing_file = external_missing_file
        
        self.test_only = []

        (fd, self.tempfilename) = tempfile.mkstemp(suffix='.nc', prefix='cfdm_', dir='.')
        os.close(fd)        
        (fd, self.tempfilename_parent) = tempfile.mkstemp(suffix='.nc', prefix='cfdm_parent_', dir='.')
        os.close(fd)        
        (fd, self.tempfilename_external) = tempfile.mkstemp(suffix='.nc', prefix='cfdm_external_', dir='.')
        os.close(fd)        
    #--- End: def

    def tearDown(self):
        os.remove(self.tempfilename)
        os.remove(self.tempfilename_parent)
        os.remove(self.tempfilename_external)
    #--- End: def

    def test_EXTERNAL_READ(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # Read the parent file on its own, without the external file
        f = cfdm.read(self.parent_file)
        self.assertTrue(len(f) == 1)
        f = f[0]

        cell_measure = f.get_construct('measure%area')

        self.assertTrue(cell_measure.nc_external())
        self.assertTrue(cell_measure.nc_get_variable() == 'areacella')
        self.assertTrue(cell_measure.properties() == {})
        self.assertFalse(cell_measure.has_data())
        
        # External file contains only the cell measure variable
        f = cfdm.read(self.parent_file, external=[self.external_file],
                      verbose=False)

        c = cfdm.read(self.combined_file, verbose=False)

        cell_measure = f[0].get_construct('measure%area')

#        print ('\nParent + External:\n')
#        for x in f:
#            print(x)
#            print (x.get_read_report())
#
#        print ('\nCombined:\n')
#        for x in c:
#            print(x)

        self.assertTrue(len(f) == 1)
        self.assertTrue(len(c) == 1)

        for i in range(len(f)):
            self.assertTrue(c[i].equals(f[i], verbose=True))

        # External file contains other variables
        f = cfdm.read(self.parent_file, external=self.combined_file,
                      verbose=False)

#        print ('\nParent + Combined:\n')
#        for x in f:
#            print(x)
#
#        print ('\nCombined:\n')
#        for x in c:
#            print(x)

        self.assertTrue(len(f) == 1)
        self.assertTrue(len(c) == 1)

        for i in range(len(f)):
            self.assertTrue(c[i].equals(f[i], verbose=True))

        # Two external files
        f = cfdm.read(self.parent_file,
                      external=[self.external_file, self.external_missing_file],
                      verbose=False)

#        print ('\nParent + External + External Missing:\n')
#        for x in f:
#            print(x)
#
#        print ('\nCombined:\n')
#        for x in c:
#            print(x)

        self.assertTrue(len(f) == 1)
        self.assertTrue(len(c) == 1)

        for i in range(len(f)):
            self.assertTrue(c[i].equals(f[i], verbose=True))
    #--- End: def        
   
    def test_EXTERNAL_WRITE(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        parent   = cfdm.read(self.parent_file)
        combined = cfdm.read(self.combined_file)
        
        # External file contains only the cell measure variable
        f = cfdm.read(self.parent_file, external=self.external_file)

        cfdm.write(f, self.tempfilename)
        g = cfdm.read(self.tempfilename)

        self.assertTrue(len(g) == len(combined))

        for i in range(len(g)):
            self.assertTrue(combined[i].equals(g[i], verbose=True))

        cell_measure = g[0].get_construct('measure%area')

        self.assertFalse(cell_measure.nc_external())
        cell_measure.nc_external(True)
        self.assertTrue(cell_measure.nc_external())
        self.assertTrue(cell_measure.properties())
        self.assertTrue(cell_measure.has_data())

        self.assertTrue(g[0].get_construct('measure%area').nc_external())

        cfdm.write(g, self.tempfilename_parent,
                   external=self.tempfilename_external,
                   verbose=False)

        h = cfdm.read(self.tempfilename_parent, verbose=False)

        self.assertTrue(len(h) == len(parent))

        for i in range(len(h)):
            self.assertTrue(parent[i].equals(h[i], verbose=True))

        h = cfdm.read(self.tempfilename_external)
        external = cfdm.read(self.external_file)

        self.assertTrue(len(h) == len(external))

        for i in range(len(h)):
            self.assertTrue(external[i].equals(h[i], verbose=True))
    #--- End: def        
    
#--- End: class


if __name__ == '__main__':
    print('Run date:', datetime.datetime.utcnow())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)




from builtins import super

from . import abstract

from . import CoordinateConversion
from . import Datum


class CoordinateReference(abstract.Container):
    '''A coordinate reference construct of the CF data model. 

A coordinate reference construct relates the coordinate values of the
coordinate system to locations in a planetary reference frame.

The domain of a field construct may contain various coordinate
systems, each of which is constructed from a subset of the dimension
and auxiliary coordinate constructs. For example, the domain of a
four-dimensional field construct may contain horizontal (y-x),
vertical (z), and temporal (t) coordinate systems. There may be more
than one of each of these, if there is more than one coordinate
construct applying to a particular spatiotemporal dimension (for
example, there could be both latitude-longitude and y-x projection
coordinate systems). In general, a coordinate system may be
constructed implicitly from any subset of the coordinate constructs,
yet a coordinate construct does not need to be explicitly or
exclusively associated with any coordinate system.

A coordinate system of the field construct can be explicitly defined
by a coordinate reference construct which relates the coordinate
values of the coordinate system to locations in a planetary reference
frame and consists of the following:

* References to the dimension coordinate and auxiliary coordinate
  constructs that define the coordinate system to which the coordinate
  reference construct applies. Note that the coordinate values are not
  relevant to the coordinate reference construct, only their
  properties.

..

* A definition of a datum specifying the zeroes of the dimension and
  auxiliary coordinate constructs which define the coordinate
  system. The datum may be implied by the metadata of the referenced
  dimension and auxiliary coordinate constructs, or explicitly
  provided.

..

* A coordinate conversion, which defines a formula for converting
  coordinate values taken from the dimension or auxiliary coordinate
  constructs to a different coordinate system. A coordinate
  reference construct relates the coordinate values of the field to
  locations in a planetary reference frame.

.. versionadded:: 1.7.0

    '''
    
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._CoordinateConversion = CoordinateConversion
        instance._Datum                = Datum
        return instance
    #--- End: def

    def __init__(self, coordinates=None, datum=None,
                 coordinate_conversion=None, source=None, copy=True):
        '''**Initialization**

:Parameters:

    coordinates: sequence of `str`, optional
        Identify the related dimension and auxiliary coordinate
        constructs by their construct identifiers. Ignored if the
        *source* parameter is set.

        *Parameter example:*
          ``coordinates=['dimensioncoordinate2']``

        *Parameter example:*
          ``coordinates=('dimensioncoordinate0', 'dimensioncoordinate1')``

        The coordinates may also be set after initialisation with the
        `coordinates` and `set_coordinate` methods.

    datum: `Datum`, optional
        Set the datum component of the coordinate reference
        construct. Ignored if the *source* parameter is set.

        The datum may also be set after initialisation with the
        `set_datum` method.

    coordinate_conversion: `CoordinateConversion`, optional
        Set the coordinate conversion component of the coordinate
        reference construct. Ignored if the *source* parameter is set.

        The coordinate conversion may also be set after initialisation
        with the `set_coordinate_conversion` method.

    source: optional
        Initialize the coordinates, datum and coordinate conversion
        from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy arguments prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__()

        self._set_component('coordinates', set(), copy=False)
        
        if source:
            try:
                coordinates = source.coordinates()
            except AttributeError:
                coordinates = None

            try:
                coordinate_conversion = source.get_coordinate_conversion()
            except AttributeError:
                coordinate_conversion = None

            try:
                datum = source.get_datum()
            except AttributeError:
                datum = None
        #--- End: if
              
        self.coordinates(coordinates)
        self.set_coordinate_conversion(coordinate_conversion, copy=copy)
        self.set_datum(datum, copy=copy)
    #--- End: def

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def construct_type(self):
        '''Return a description of the construct type.
        
.. versionadded:: 1.7.0
        
:Returns:

    `str`
        The construct type.

        '''
        return 'coordinate_reference'

    @property
    def coordinate_conversion(self):
        '''Return the coordinate conversion component.

.. versionadded:: 1.7.0

.. seealso:: `datum`, `get_coordinate_conversion`

:Returns:

    `CoordinateConversion`
        The coordinate conversion.

**Examples:**

>>> c.coordinate_conversion
<>
        '''
        return self.get_coordinate_conversion()
    #--- End: def
        
    @property
    def datum(self):
        '''Return the datum component.

.. versionadded:: 1.7.0

.. seealso:: `coordinate_conversion`, `datum`

:Returns:

   `Datum`
        The datum.

**Examples:**

>>> c.datum
<>
        '''
        return self.get_datum()
    #--- End: def

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def coordinates(self, coordinates=None):
        '''Return or replace all references to coordinate constructs.

.. seealso:: `del_coordinate`, `set_coordinate`

:Parameters:

    coordinates: sequence of `str`

:Returns:

    `set`
        The identifiers of the coordinate objects.

**Examples:**

>>> c.coordinates()
{'dimensioncoordinate0',
 'dimensioncoordinate1',
 'auxiliarycoordinate0',
 'auxiliarycoordinate1'}

        '''
        out = self._get_component('coordinates').copy()

        if coordinates is not None:
            self._set_component('coordinates', set(coordinates), copy=False)

        return out
    #--- End: def
            
    def copy(self):
        '''Return a deep copy.

``f.copy()`` is equivalent to ``copy.deepcopy(f)``.

.. versionadded:: 1.7.0

:Returns:

        The deep copy.

**Examples:**

>>> d = c.copy()

        '''
        return type(self)(source=self, copy=True)
    #--- End: def

    def del_coordinate(self, cid, *default):
        '''Remove a reference to a coordinate construct.

.. versionadded:: 1.7.0

.. seealso:: `coordinates`, `set_coordinate`

:Parameters:

    cid: `str`
        The construct identifier of the coordinate construct.

          *Parameter example:*
             ``cid='dimensioncoordinate1'``

          *Parameter example:*
             ``cid='auxiliarycoordinate0'``

    default: optional
        Return *default* if the coordinate construct has not been
        referenced.

:Returns:

      The removed coordinate construct identifier property. If unset
      then *default* is returned, if provided.

**Examples:**

>>> c.coordinates()
{'dimensioncoordinate0',
 'dimensioncoordinate1'}
>>> c.del_coordinate('dimensioncoordinate0')
'dimensioncoordinate0'
>>> c.coordinates()
{'dimensioncoordinate1'}
>>> c.del_coordinate('dimensioncoordinate0', 'not set')
'not set'

        '''
        coordinates = self._get_component('coordinates')
        if cid in coordinates:
            coordinates.remove(cid)
            return cid

        if default:
            return default[0]
        
        raise AttributeError("{!r} has no {!r} coordinate".format(
            self.__class__.__name__, cid))
    #--- End: def
    
    def del_coordinate_conversion(self):
        '''Remove the coordinate conversion component.

.. versionadded:: 1.7.0

.. seealso:: `coordinate_conversion`, `get_coordinate_conversion`,
             `set_coordinate_conversion`

:Returns:

        The removed coordinate conversion component.

**Examples:**

>>> c.del_coordinate_conversion()
<CoordinateConversion: Parameters(grid_mapping_name, grid_north_pole_latitude, grid_north_pole_longitude)>
>>> c.get_coordinate_conversion()
<CoordinateConversion: >

        '''
        new = self._CoordinateConversion()
        out = self._del_component('coordinate_conversion', new)
        self.set_coordinate_conversion(new)
        return out
    #--- End: def
    
    def del_datum(self):
        '''Remove the datum component.

.. versionadded:: 1.7.0

.. seealso:: `datum`, `get_datum`, `set_datum`

:Returns:

        The removed datum component.

**Examples:**

>>> c.del_datum()
<Datum: Parameters(earth_radius)>
>>> c.get_datum()
<Datum: >

        '''
        new = self._Datum()
        out = self._del_component('datum', new)
        self.set_datum(new)
        return out
    #--- End: def

    def get_coordinate_conversion(self):
        '''Get the coordinate conversion component.

.. versionadded:: 1.7.0

.. seealso:: `coordinate_conversion`, `del_coordinate_conversion`,
             `set_coordinate_conversion`

:Returns:

    `CoordinateConversion`
        The coordinate conversion component.

**Examples:**

>>> c.get_coordinate_conversion()
<CoordinateConversion: Parameters(grid_mapping_name, grid_north_pole_latitude, grid_north_pole_longitude)>

        '''
        out = self._get_component('coordinate_conversion', None)
        if out is None:
            out = self._CoordinateConversion()
            self.set_coordinate_conversion(out)

        return out
    #--- End: def
    
    def get_datum(self):
        '''Return the datum component.

.. versionadded:: 1.7.0

.. seealso:: `datum`, `get_coordinate_conversion`

:Returns:

   `Datum`
        The datum component.

**Examples:**

>>> c.get_datum()
<Datum: Parameters(earth_radius)>

        '''
        out = self._get_component('datum', None)
        if out is None:
            out = self._Datum()
            self.set_datum(out)
            
        return out
    #--- End: def
    
    def set_coordinate(self, cid):
        '''Set a reference to a coordinate construct.

.. versionadded:: 1.7.0

.. seealso:: `coordinates`, `del_coordinate`

:Parameters:

    cid: `str`
        The construct identifier of the coordinate construct.

          *Parameter example:*
             ``cid='dimensioncoordinate1'``

          *Parameter example:*
             ``cid='auxiliarycoordinate0'``

:Returns:

    `None`

**Examples:**

>>> c.coordinates()
{'dimensioncoordinate0',
 'dimensioncoordinate1'}
>>> c.set_coordinates('auxiliarycoordinate0')
>>> c.coordinates()
{'dimensioncoordinate0',
 'dimensioncoordinate1',
 'auxiliarycoordinate0'}

        '''
        c = self._get_component('coordinates')
        c.add(cid)
    #--- End: def

    def set_coordinate_conversion(self, coordinate_conversion, copy=True):
        '''Set the coordinate conversion component.

.. versionadded:: 1.7.0

.. seealso:: `coordinate_conversion`, `del_coordinate_conversion`,
             `get_coordinate_conversion`

:Parameters:

    coordinate_conversion: `CoordinateConversion`
        The coordinate conversion component to be inserted.

    copy: `bool`, optional
        If False then do not copy the coordinate conversion prior to
        insertion. By default the coordinate conversion is copied.

:Returns:

    `None`

**Examples:**

>>> c.set_coordinate_conversion(cc)

>>> c.set_coordinate_conversion(cc, copy=False)

        '''
        if copy and coordinate_conversion is not None:
            coordinate_conversion = coordinate_conversion.copy()
            
        self._set_component('coordinate_conversion',
                            coordinate_conversion, copy=False)
    #--- End: def

    def set_datum(self, datum, copy=True):
        '''Set the datum component.

.. versionadded:: 1.7.0

.. seealso:: `datum`, `del_datum`, `get_datum`

:Parameters:

    datum: `Datum`
        The datum component to be inserted.

    copy: `bool`, optional
        If False then do not copy the datum prior to insertion. By
        default the datum is copied.

:Returns:

    `None`

**Examples:**

>>> c.set_datum(d)

>>> c.set_datum(d, copy=False)

        '''
        if copy and datum is not None:
            datum = datum.copy()
            
        self._set_component('datum', datum, copy=False)
    #--- End: def

#--- End: class



from future.utils import with_metaclass
from builtins import (str, super)

import abc

from copy import deepcopy

from . import Container


class Properties(with_metaclass(abc.ABCMeta, Container)):
    '''Abstract base class for an object with descriptive properties.

.. versionadded:: 1.7.0

    '''
    def __init__(self, properties=None, source=None, copy=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
           ``properties={'standard_name': 'altitude'}``
        
        Properties may also be set after initialisation with the
        `properties` and `set_property` methods.

    source: optional
        Initialize the properties from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization By default parameters are deep copied.

        '''
        super().__init__()

        self._set_component('properties', {}, copy=False)
        
        if source is not None:
            try:
                properties = source.properties()
            except AttributeError:
                properties = None
        #--- End: if
        
        if properties:
            self.properties(properties, copy=copy)
    #--- End: def

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self):
        '''Return a deep copy.

``f.copy()`` is equivalent to ``copy.deepcopy(f)``.

.. versionadded:: 1.7.0

:Returns:

        The deep copy.

**Examples:**

>>> g = f.copy()

        '''
        return type(self)(source=self, copy=True)
    #--- End: def

    def del_property(self, prop, *default):
        '''Remove a property.

.. versionadded:: 1.7.0

.. seealso:: `get_property`, `has_property`, `properties`,
             `set_property`

:Parameters:

    prop: `str`
        The name of the property to be removed.

        *Parameter example:*
           ``prop='long_name'``

    default: optional
        Return *default* if the property has not been set.

:Returns:

        The removed property. If unset then *default* is returned, if
        provided.

**Examples:**

>>> f.set_property('project', 'CMIP7')
>>> f.has_property('project')
True
>>> f.get_property('project')
'CMIP7'
>>> f.del_property('project')
'CMIP7'
>>> f.has_property('project')
False
>>> print(f.del_property('project', None))
None
>>> print(f.get_property('project', None))
None

        '''
        try:
            return self._get_component('properties').pop(prop)
        except KeyError:
            if default:
                return default[0]

            raise AttributeError("{!r} has no {!r} property".format(
                self.__class__.__name__, prop))
    #--- End: def

    def get_property(self, prop, *default):
        '''Return a property.

.. versionadded:: 1.7.0

.. seealso:: `del_property`, `has_property`, `properties`,
             `set_property`

:Parameters:

    prop: `str`
        The name of the property to be returned.

        *Parameter example:*
           ``prop='standard_name'``

    default: optional
        Return *default* if the property has not been set.

:Returns:

        The value of the property. If unset then *default* is
        returned, if provided.

**Examples:**

>>> f.set_property('project', 'CMIP7')
>>> f.has_property('project')
True
>>> f.get_property('project')
'CMIP7'
>>> f.del_property('project')
'CMIP7'
>>> f.has_property('project')
False
>>> print(f.del_property('project', None))
None
>>> print(f.get_property('project', None))
None

        '''
        try:
            return self._get_component('properties')[prop]
        except KeyError:
            if default:
                return default[0]

            raise AttributeError("{!r} has no {!r} property".format(
                self.__class__.__name__, prop))
    #--- End: def

    def has_property(self, prop):
        '''Whether a property has been set.

.. versionadded:: 1.7.0

.. seealso:: `del_property`, `get_property`, `properties`,
             `set_property`

:Parameters:

    prop: `str`
        The name of the property.

        *Parameter example:*
           ``prop='long_name'``

:Returns:

     `bool`
        True if the property has been set, otherwise False.

**Examples:**

>>> f.set_property('project', 'CMIP7')
>>> f.has_property('project')
True
>>> f.get_property('project')
'CMIP7'
>>> f.del_property('project')
'CMIP7'
>>> f.has_property('project')
False
>>> print(f.del_property('project', None))
None
>>> print(f.get_property('project', None))
None

        '''
        return prop in self._get_component('properties')
    #--- End: def

    def properties(self, properties=None, copy=True):
        '''Return or replace all properties.

.. versionadded:: 1.7.0

.. seealso:: `del_property`, `get_property`, `has_property`,
             `set_property`

:Parameters:

    properties: `dict`, optional   
        Delete all existing properties, and instead store the
        properties from the dictionary supplied.


        *Parameter example:*
          ``properties={'standard_name': 'altitude', 'foo': 'bar'}``
        
        *Parameter example:*
          ``properties={}``        

    copy: `bool`, optional
        If False then any property values provided by the *properties*
        parameter are not copied before insertion. By default they are
        deep copied.

:Returns:

    `dict`
        The properties or, if the *properties* parameter was set, the
        original properties.

**Examples:**

>>> p = f.properties({'standard_name': 'altitude', 'foo': 'bar'})
>>> f.properties()
{'standard_name': 'altitude',
 'foo': 'bar'}
>>> f.properties({})
{'standard_name': 'altitude',
 'foo': 'bar'}
>>> f.properties()
{}

        '''
        out = self._get_component('properties').copy()

        if properties is not None:
            if copy:
                properties = deepcopy(properties)                
            else:
                properties = properties.copy()

            self._set_component('properties', properties, copy=False)

        return out
    #--- End: def

    def set_property(self, prop, value, copy=True):
        '''Set a property.

.. versionadded:: 1.7.0

.. seealso:: `del_property`, `get_property`, `has_property`,
             `properties`

:Parameters:

    prop: `str`
        The name of the property to be set.

    value:
        The value for the property.

    copy: `bool`, optional
        If True then set a deep copy of *value*.

:Returns:

     `None`

**Examples:**

>>> f.set_property('project', 'CMIP7')
>>> f.has_property('project')
True
>>> f.get_property('project')
'CMIP7'
>>> f.del_property('project')
'CMIP7'
>>> f.has_property('project')
False
>>> print(f.del_property('project', None))
None
>>> print(f.get_property('project', None))
None

        '''
        if copy:
            value = deepcopy(value)
            
        self._get_component('properties')[prop] = value
    #--- End: def

#--- End: class



import sys

import numpy

CONSTANTS = {
    # The value of absolute tolerance for testing numerically tolerant
    # equality.
    'RTOL': sys.float_info.epsilon,
    # The value of relative tolerance for testing numerically tolerant
    # equality.
    'ATOL': sys.float_info.epsilon,
}
masked = numpy.ma.masked



from __future__ import print_function
from builtins import super

import textwrap

import numpy
import sys

from . import Container


class Properties(Container):
    '''Mixin class for descriptive properties.

.. versionadded:: 1.7.0

    '''
    def _dump_properties(self, _prefix='', _level=0,
                         _omit_properties=None):
        '''TODO

.. versionadded:: 1.7.0

:Parameters:

    omit: sequence of `str`, optional
        Omit the given CF properties from the description.

    _level: `int`, optional

:Returns:

    `str`

**Examples:**

'''
        indent0 = '    ' * _level
        string = []

        properties = self.properties()
        
        if _omit_properties:
            for prop in _omit_properties:
                 properties.pop(prop, None)
        #--- End: if
 
        for prop, value in sorted(properties.items()):
            name   = '{0}{1}{2} = '.format(indent0, _prefix, prop)
            value  = repr(value)
            subsequent_indent = ' ' * len(name)
            if value.startswith("'") or value.startswith('"'):
                subsequent_indent = '{0} '.format(subsequent_indent)
                
            string.append(
                textwrap.fill(name+value, 79,
                              subsequent_indent=subsequent_indent))
        
        return '\n'.join(string)
    #--- End: def

    def equals(self, other, rtol=None, atol=None, verbose=False,
               ignore_data_type=False, ignore_fill_value=False,
               ignore_properties=(), ignore_type=False):
        '''Whether two instances are the same.

Equality is strict by default. This means that:

* the same descriptive properties must be present, with the same
  values and data types, and vector-valued properties must also have
  same the size and be element-wise equal (see the *ignore_properties*
  and *ignore_data_type* parameters).

Two real numbers ``x`` and ``y`` are considered equal if
``|x-y|<=atol+rtol|y|``, where ``atol`` (the tolerance on absolute
differences) and ``rtol`` (the tolerance on relative differences) are
positive, typically very small numbers. See the *atol* and *rtol*
parameters.

Any type of object may be tested but, in general, equality is only
possible with another object of the same type, or a subclass of
one. See the *ignore_type* parameter.

.. versionadded:: 1.7.0

:Parameters:

    other: 
        The object to compare for equality.

    atol: float, optional
        The tolerance on absolute differences between real
        numbers. The default value is set by the `cfdm.ATOL` function.
        
    rtol: float, optional
        The tolerance on relative differences between real
        numbers. The default value is set by the `cfdm.RTOL` function.

    ignore_fill_value: `bool`, optional
        If True then the "_FillValue" and "missing_value" properties
        are omitted from the comparison.

    verbose: `bool`, optional
        If True then print information about differences that lead to
        inequality.

    ignore_properties: sequence of `str`, optional
        The names of properties to omit from the comparison.

    ignore_data_type: `bool`, optional
        If True then ignore the data types in all numerical
        comparisons. By default different numerical data types imply
        inequality, regardless of whether the elements are within the
        tolerance for equality.

    ignore_type: `bool`, optional
        Any type of object may be tested but, in general, equality is
        only possible with another object of the same type, or a
        subclass of one. If *ignore_type* is True then equality is
        possible for any object with a compatible API.

:Returns: 
  
    `bool`
        Whether the two instances are equal.

**Examples:**

>>> p.equals(p)
True
>>> p.equals(p.copy())
True
>>> p.equals('not a colection of properties')
False

>>> q = p.copy()
>>> q.set_property('foo', 'bar')
>>> p.equals(q)
False
>>> p.equals(q, verbose=True)
Field: Non-common property name: foo
Field: Different properties
False

        '''
        pp = super()._equals_preprocess(other, verbose=verbose,
                                        ignore_type=ignore_type)
        if pp in (True, False):
            return pp
        
        other = pp
        
        # ------------------------------------------------------------
        # Check the properties
        # ------------------------------------------------------------
        if ignore_fill_value:
            ignore_properties += ('_FillValue', 'missing_value')

        self_properties  = self.properties()
        other_properties = other.properties()

        if ignore_properties:
            for prop in ignore_properties:
                self_properties.pop(prop, None)
                other_properties.pop(prop, None)
        #--- End: if
                
        if set(self_properties) != set(other_properties):
            if verbose:
                _ =  set(self_properties).symmetric_difference(other_properties)
                for prop in set(self_properties).symmetric_difference(other_properties):                    
                    print("{0}: Non-common property name: {1}".format( 
                        self.__class__.__name__, prop))
            return False

        for prop, x in self_properties.items():
            y = other_properties[prop]

            if not self._equals(x, y,
                                rtol=rtol, atol=atol,
                                ignore_fill_value=ignore_fill_value,
                                ignore_data_type=True, #ignore_data_type,
                                verbose=verbose):
                if verbose:
                    print("{0}: Different {1}: {2!r}, {3!r}".format(
                        self.__class__.__name__, prop, x, y))
                return False
        #--- End: for

        return True
    #--- End: def

#--- End: class



from __future__ import print_function
import datetime
import inspect
import unittest

import cfdm

class FunctionsTest(unittest.TestCase):
    def setUp(self):
        self.test_only = []
    #--- End: def
        
    def test_CF(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        self.assertTrue(cfdm.CF() == '1.7')
    #--- End: def
    
#--- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    cfdm.environment()
    print('')
    unittest.main(verbosity=2)



from builtins import object
#from future.utils import with_metaclass

#import abc


#class ConstructAccess(with_metaclass(abc.ABCMeta, object)):
class ConstructAccess(object):
    '''Mixin class for accessing an embedded `Constructs` object.

.. versionadded:: 1.7.0

    '''   
#    @abc.abstractmethod
    def _get_constructs(self, *default):
        '''Return the `Constructs` object

.. versionadded:: 1.7.0

:Parameters:

    default: optional
        If set then return *default* if there is no `Constructs` object.

:Returns:

        The `Constructs` object. If unset then return *default* if provided.

**Examples:**

>>> c = f._get_constructs()
>>> c = f._get_constructs(None)

        '''
        raise NotImplementedError("Must implement this method in subclasses")
    #--- End: def
    
    def data_constructs(self, copy=False):
        '''Return metadata constructs that support data arrays.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `get_construct`

:Parameters:

    copy: `bool`, optional
        If True then return copies of the constructs. By default the
        constructs are not copied.

:Returns:

    `dict`
        Constructs are returned as values of a dictionary, keyed by
        their construct identifiers.

**Examples:**

>>> f.constructs()
TODO
>>> f.data_constructs()
TODO

        '''
        return self._get_constructs().data_constructs(copy=copy)
    #-- End: def
    
    def constructs_data_axes(self):
        '''Return the domain axes spanned by metadata construct data arrays.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `get_construct`

:Returns:

    `dict`

**Examples:**

>>> f.constructs_data_axes()
TODO

        '''
        return self._get_constructs().constructs_data_axes()
    #--- End: def
    
    def construct_type(self, key):
        '''TODO

.. versionadded:: 1.7.0
        '''                
        return self._get_constructs().construct_type(key)
    #--- End: def
    
    def constructs(self, construct_type=None, copy=False):
        '''Return metadata constructs.

.. versionadded:: 1.7.0

.. seealso:: `constructs_data_axes`, `del_construct`, `get_construct`,
             `get_construct_key`, `has_construct`, `set_construct`

:Parameters:

    construct_type: (sequence of) `str`, optional
        Select constructs of the given type, or types. Valid types
        are:

          ==========================  ================================
          *construct_type*            Constructs
          ==========================  ================================
          ``'domain_ancillary'``      Domain ancillary constructs
          ``'dimension_coordinate'``  Dimension coordinate constructs
          ``'domain_axis'``           Domain axis constructs
          ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
          ``'cell_measure'``          Cell measure constructs
          ``'coordinate_reference'``  Coordinate reference constructs
          ``'cell_method'``           Cell method constructs
          ``'field_ancillary'``       Field ancillary constructs
          ==========================  ================================

        *Parameter example:*
          ``construct_type='dimension_coordinate'``

        *Parameter example:*
          ``construct_type=['auxiliary_coordinate']``

        *Parameter example:*
          ``construct_type=('domain_ancillary', 'cell_method')``

        Note that a domain never contains cell method nor field
        ancillary constructs.

    copy: `bool`, optional
        If `True` then deep copies of the constructs are returned.

:Returns:

    `dict`
        Constructs are returned as values of a dictionary, keyed by
        their construct identifiers.
        
        If cell method contructs, and no other construct types, have
        been selected with the *construct_type* parameter then the
        constructs are returned in an ordered dictionary
        (`collections.OrderedDict`). The order is determined by the
        order in which the cell method constructs were originally
        added.

**Examples:**

>>> f.constructs()
{}

>>> f.constructs()
{'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
 'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
 'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name:Grid latitude name(10) >,
 'cellmeasure0': <CellMeasure: measure%area(9, 10) km2>,
 'cellmethod0': <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>,
 'cellmethod1': <CellMethod: domainaxis3: maximum>,
 'coordinatereference0': <CoordinateReference: atmosphere_hybrid_height_coordinate>,
 'coordinatereference1': <CoordinateReference: rotated_latitude_longitude>,
 'dimensioncoordinate0': <DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >,
 'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>,
 'dimensioncoordinate2': <DimensionCoordinate: grid_longitude(9) degrees>,
 'dimensioncoordinate3': <DimensionCoordinate: time(1) days since 2018-12-01 >,
 'domainancillary0': <DomainAncillary: ncvar%a(1) m>,
 'domainancillary1': <DomainAncillary: ncvar%b(1) >,
 'domainancillary2': <DomainAncillary: surface_altitude(10, 9) m>,
 'domainaxis0': <DomainAxis: 1>,
 'domainaxis1': <DomainAxis: 10>,
 'domainaxis2': <DomainAxis: 9>,
 'domainaxis3': <DomainAxis: 1>,
 'fieldancillary0': <FieldAncillary: air_temperature standard_error(10, 9) K>}
>>> f.constructs(construct_type='coordinate_reference')
{'coordinatereference0': <CoordinateReference: atmosphere_hybrid_height_coordinate>,
 'coordinatereference1': <CoordinateReference: rotated_latitude_longitude>}
>>> f.constructs(construct_type='cell_method')
OrderedDict([('cellmethod0', <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>),
             ('cellmethod1', <CellMethod: domainaxis3: maximum>)])
>>> f.constructs(construct_type=['cell_method', 'field_ancillary'])
{'cellmethod0': <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>,
 'cellmethod1': <CellMethod: domainaxis3: maximum>,
 'fieldancillary0': <FieldAncillary: air_temperature standard_error(10, 9) K>}

        '''
        return self._get_constructs().constructs(construct_type=construct_type,
                                                 copy=copy)
    #--- End: def
    
    def del_construct(self, key):
        '''Remove a metadata construct.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `get_construct`, `has_construct`,
             `set_construct`

:Parameters:

    key: `str`, optional
        The identifier of the metadata construct.

        *Parameter example:*
          ``key='auxiliarycoordinate0'``
        
:Returns:

        The removed metadata construct.

**Examples:**

>>> f.del_construct('dimensioncoordinate1')
<DimensionCoordinate: grid_latitude(111) degrees>

        '''
        return self._get_constructs().del_construct(key)
    #--- End: def

    def get_construct(self, key):
        '''Return a metadata construct.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `del_construct`, `has_construct`,
             `set_construct`

:Parameters:

    key: `str`
        The identifier of the metadata construct.

        *Parameter example:*
          ``key='domainaxis1'``

:Returns:

        The metadata construct.

**Examples:**

>>> f.constructs()
{'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degree_N>,
 'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degreeE>,
 'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name:greek_letters(10) >,
 'coordinatereference1': <CoordinateReference: rotated_latitude_longitude>,
 'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>,
 'dimensioncoordinate2': <DimensionCoordinate: grid_longitude(9) degrees>,
 'domainaxis1': <DomainAxis: 10>,
 'domainaxis2': <DomainAxis: 9>}
>>> f.get_construct('dimensioncoordinate1')
<DimensionCoordinate: grid_latitude(10) degrees>

        '''
        return self._get_constructs().get_construct(key)
    #--- End: def

    def get_construct_data_axes(self, key, *default):
        '''Return the domain axis identifiers spanned by a metadata construct
data array.

.. versionadded:: 1.7.0

.. seealso:: `constructs_data_axes`, `get_construct`,
             `set_construct_data_axes`

:Parameters:

    key: `str`
        The construct identifier of the metadata construct.

        *Parameter example:*
          ``key='domainancillary1'``

    default: optional
        Return *default* if the metadata construct does not have a
        data array, or no domain axes have been set.

:Returns:

        The identifiers of the domain axis constructs spanned by data
        array. If a metadata construct does not have a data array, or
        no domain axes have been set, then `None` is returned.

**Examples:**

>>> f.get_construct_data_axes('domainancillary2')
('domainaxis1', 'domainaxis0')
>>> da = f.get_construct('domainancillary2')
>>> data = da.del_data()
>>> print(f.get_construct_data_axes('domainancillary2', None))
None

>>> f.get_construct_data_axes('cellmethod0', 'no axes')
'no axes'

        '''
        axes = self.constructs_data_axes().get(key)
        if axes is None:
            if default:
                return default[0]

            raise ValueError("no axes.")

        return axes
    #--- End: def
    
    def has_construct(self, key):
        '''Whether a construct exists.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `del_construct`, `get_construct`,
             `set_construct`

:Parameters:

    key: `str`
        The identifier of the metadata construct.

        *Parameter example:*
          ``key='cellmeasure1'``

:Returns:

    `bool`
        True if the metadata construct exists, otherwise False.

**Examples:**

>>> f.constructs()
{'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degree_N>,
 'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degreeE>,
 'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name:greek_letters(10) >,
 'coordinatereference1': <CoordinateReference: rotated_latitude_longitude>,
 'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>,
 'dimensioncoordinate2': <DimensionCoordinate: grid_longitude(9) degrees>,
 'domainaxis1': <DomainAxis: 10>,
 'domainaxis2': <DomainAxis: 9>}
>>> f.has_construct('dimensioncoordinate1')
True
>>> f.has_construct('domainaxis99')
False

        '''
        return self._get_constructs().has_construct(key)
    #--- End: def

    def domain_axis_name(self, axis):
        '''TODO WHY DO WE NED THIS HERE?

.. versionadded:: 1.7.0

        '''
        return self._get_constructs().domain_axis_name(axis)
    #--- End: def
    
    def set_construct(self, construct, key=None, axes=None,
                      copy=True):
        '''Set a metadata construct.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `del_construct`, `get_construct`,
             `set_construct_data_axes`

:Parameters:

    construct:
        The metadata construct to be inserted.

    key: `str`, optional
        The construct identifier to be used for the construct. If not
        set then a new, unique identifier is created automatically. If
        the identifier already exisits then the exisiting construct
        will be replaced.

        *Parameter example:*
          ``key='cellmeasure0'``

    axes: (sequence of) `str`, optional
        The construct identifiers of the domain axis constructs
        spanned by the data array. An exception is raised if used for
        a metadata construct that can not have a data array,
        i.e. domain axis, cell method and coordinate reference
        constructs.

        *Parameter example:*
          ``axes='domainaxis1'``
        
        *Parameter example:*
          ``axes=['domainaxis1']``
        
        *Parameter example:*
          ``axes=['domainaxis1', 'domainaxis0']``
        
    copy: `bool`, optional
        If True then return a copy of the unique selected
        construct. By default the construct is not copied.
        
:Returns:

     `str`
        The construct identifier for the construct.
    
**Examples:**

>>> key = f.set_construct(c)
>>> key = f.set_construct(c, copy=False)
>>> key = f.set_construct(c, axes='domainaxis2')
>>> key = f.set_construct(c, key='cellmeasure0')

        '''
        return self._get_constructs().set_construct(construct, key=key,
                                                    axes=axes, copy=copy)
    #--- End: def

    def set_construct_data_axes(self, key, axes):
        '''Set the domain axis constructs spanned by a metadata construct data
array.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `del_construct`, `get_construct`,
             `set_construct`

:Parameters:

    key: `str`
        The construct identifier of the metadata construct.

        *Parameter example:*
          ``key='dimensioncoordinate2'``

     axes: (sequence of) `str`
        The construct identifiers of the domain axis constructs
        spanned by the data array.

        The axes may also be set with the `set_construct` method.

        *Parameter example:*
          ``axes='domainaxis1'``

        *Parameter example:*
          ``axes=['domainaxis1']``

        *Parameter example:*
          ``axes=['domainaxis1', 'domainaxis0']``

:Returns:

    `None`

**Examples:**

>>> key = f.set_construct(c)
>>> f.set_construct_data_axes(key, axes='domainaxis1')

        '''
        return self._get_constructs().set_construct_data_axes(key, axes)
    #--- End: def

#--- End: class



from __future__ import print_function
from builtins import super

from . import mixin
from . import core


class DomainAxis(mixin.NetCDFDimension,
                 mixin.Container,
                 core.DomainAxis):
    '''A domain axis construct of the CF data model. 

A domain axis construct specifies the number of points along an
independent axis of the domain. It comprises a positive integer
representing the size of the axis. In CF-netCDF it is usually defined
either by a netCDF dimension or by a scalar coordinate variable, which
implies a domain axis of size one. The field construct's data array
spans the domain axis constructs of the domain, with the optional
exception of size one axes, because their presence makes no difference
to the order of the elements.

.. versionadded:: 1.7.0

    '''
    def __init__(self, size=None, source=None, copy=True):
        '''**Initialization**

:Parameters:

    size: `int`, optional
        The size of the domain axis.

        *Parameter example:*
          ``size=192``

        The size may also be set after initialisation with the
        `set_size` method.

    source: optional
        Initialize the size from that of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(size=size, source=source, copy=copy)
        
        self._initialise_netcdf(source)
    #--- End: def
        
    def __str__(self):
        '''Called by the `str` built-in function.

x.__str__() <==> str(x)

.. versionadded:: 1.7.0

        '''
        return str(self.get_size(''))
    #--- End: def

    def equals(self, other, verbose=False, ignore_type=False):
        '''Whether two domain axis constructs are the same.

Equality is strict by default. This means that:

* the axis sizes must be the same.

Any type of object may be tested but, in general, equality is only
possible with another domain axis construct, or a subclass of one. See
the *ignore_type* parameter.

NetCDF elements, such as netCDF variable and dimension names, do not
constitute part of the CF data model and so are not checked.

.. versionadded:: 1.7.0

:Parameters:

    other: 
        The object to compare for equality.

    verbose: `bool`, optional
        If True then print information about differences that lead to
        inequality.

    ignore_type: `bool`, optional
        Any type of object may be tested but, in general, equality is
        only possible with another domain axis construct, or a
        subclass of one. If *ignore_type* is True then 
        ``DomainAxis(source=other)`` is tested, rather than the
        ``other`` defined by the *other* parameter.

:Returns: 
  
    `bool`
        Whether the two domain axis constructs are equal.

**Examples:**

>>> d.equals(d)
True
>>> d.equals(d.copy())
True
>>> d.equals('not a domain axis')
False

>>> d = cfdm.DomainAxis(1)
>>> e = cfdm.DomainAxis(99)
>>> d.equals(e, verbose=True)
DomainAxis: Different axis sizes: 1 != 99
False

        '''
        pp = super()._equals_preprocess(other, verbose=verbose,
                                        ignore_type=ignore_type)
        if pp in (True, False):
            return pp
        
        other = pp

        # Check that each axis has the same size
        self_size  = self.get_size(None)
        other_size = other.get_size(None)
        if not self_size == other_size:
            if verbose:
                print("{0}: Different axis sizes: {1} != {2}".format(
			self.__class__.__name__, self_size, other_size))
            return False

        return True
    #--- End: def

    def name(self, default=None, ncdim=True, custom=None,
             all_names=False):
        '''Return a name for the domain axis construct.

By default the name is the first found of the following:

1. The netCDF dimension name, preceeded by 'ncdim%'.
2. The value of the default parameter.

.. versionadded:: 1.7.0

:Parameters:

    default: optional
        If no name can be found then return the value of the *default*
        parameter. By default the default is `None`.

    ncdim: `bool`, optional
        If False then do not consider the netCDF dimension name.

    all_names: `bool`, optional
        If True then return a list of all possible names.

    custom: optional
        *Ignored.*

:Returns:

        The name. If the *all_names* parameter is True then a list of
        all possible names.

**Examples:**

>>> d.name()
'ncdim%time'
>>> d.name(all_names=True)
['ncdim%time']
>>> d.name('default_value', all_names=True)
['ncdim%time', 'default_value']
>>> d.nc_del_dimension()
'time'
>>> d.name('default value')
'default value'
>>> d.name('default value', all_names=True)
['default value']

        '''
        out = []

        if all_names:
            if ncdim:
                n = self.nc_get_dimension(None)
                if n is not None:
                    out.append('ncdim%{0}'.format(n))
            #--- End: if
            
            if default is not None:
                out.append(default)

            return out

        if ncdim:
            n = self.nc_get_dimension(None)
            if n is not None:
                return 'ncdim%{0}'.format(n)
        #--- End: if
        
        return default
    #--- End: def

#--- End: class



from . import abstract


class FieldAncillary(abstract.PropertiesData):
    '''A field ancillary construct of the CF data model.

The field ancillary construct provides metadata which are distributed
over the same sampling domain as the field itself. For example, if a
data variable holds a variable retrieved from a satellite instrument,
a related ancillary data variable might provide the uncertainty
estimates for those retrievals (varying over the same spatiotemporal
domain).

The field ancillary construct consists of an array of the ancillary
data, which is zero-dimensional or which depends on one or more of the
domain axes, and properties to describe the data. It is assumed that
the data do not depend on axes of the domain which are not spanned by
the array, along which the values are implicitly propagated. CF-netCDF
ancillary data variables correspond to field ancillary
constructs. Note that a field ancillary construct is constrained by
the domain definition of the parent field construct but does not
contribute to the domain's definition, unlike, for instance, an
auxiliary coordinate construct or domain ancillary construct.

.. versionadded:: 1.7.0

    '''
    @property
    def construct_type(self):
        '''Return a description of the construct type.
        
.. versionadded:: 1.7.0

:Returns:

    `str`
        The construct type.

        '''
        return 'field_ancillary'
    #--- End: def

#--- End: class



from builtins import super

from . import abstract
from . import mixin

from . import Constructs


class Field(mixin.ConstructAccess, abstract.PropertiesData):
    '''A field construct of the CF data model.

The field construct is central to the CF data model, and includes all
the other constructs. A field corresponds to a CF-netCDF data variable
with all of its metadata. All CF-netCDF elements are mapped to a field
construct or some element of the CF field construct. The field
construct contains all the data and metadata which can be extracted
from the file using the CF conventions.

The field construct consists of a data array and the definition of its
domain (that describes the locations of each cell of the data array),
field ancillary constructs containing metadata defined over the same
domain, and cell method constructs to describe how the cell values
represent the variation of the physical quantity within the cells of
the domain. The domain is defined collectively by the following
constructs of the CF data model: domain axis, dimension coordinate,
auxiliary coordinate, cell measure, coordinate reference and domain
ancillary constructs. All of the constructs contained by the field
construct are optional.

The field construct also has optional properties to describe aspects
of the data that are independent of the domain. These correspond to
some netCDF attributes of variables (e.g. units, long_name and
standard_name), and some netCDF global file attributes (e.g. history
and institution).

.. versionadded:: 1.7.0

    '''

    # Define the base of the identity keys for each construct type
    _construct_key_base = {'auxiliary_coordinate': 'auxiliarycoordinate',
                           'cell_measure'        : 'cellmeasure',
                           'cell_method'         : 'cellmethod',
                           'coordinate_reference': 'coordinatereference',
                           'dimension_coordinate': 'dimensioncoordinate',
                           'domain_ancillary'    : 'domainancillary',
                           'domain_axis'         : 'domainaxis',
                           'field_ancillary'     : 'fieldancillary',
    }
    
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._Constructs = Constructs
        return instance
    
#        obj = object.__new__(cls, *args, **kwargs)
#        obj._Constructs = Constructs
#        return obj
    #--- End: def

    def __init__(self, properties={}, source=None, copy=True,
                 _use_data=True):
        '''**Initialization**
        
:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
           ``properties={'standard_name': 'air_temperature'}``
        
        Properties may also be set after initialisation with the
        `properties` and `set_property` methods.

    source: optional
        Initialize the properties, data and metadata constructs from
        those of *source*.
        
    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(properties=properties, source=source,
                         copy=copy, _use_data=False)

        if source is not None:
            # Initialise constructs and the data from the source
            # parameter
            try:
                constructs = source._get_constructs(None)
            except AttributeError:
                constructs = None
                
            try:
                data = source.get_data(None) 
            except AttributeError:
                data = None

            try:
                data_axes = source.get_data_axes(None)
            except AttributeError:
                data_axes = None
                                
            if constructs is not None and (copy or not _use_data):
                constructs = constructs.copy(data=_use_data)
        else:
            constructs = None
            data       = None
            data_axes  = None

        if constructs is None:
            constructs = self._Constructs(**self._construct_key_base)
            
        self._set_component('constructs', constructs, copy=False)

        if data is not None and _use_data:
            self.set_data(data, data_axes, copy=copy)
        elif data_axes is not None:
            self.set_data_axes(data_axes)
     #--- End: def
    
    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    def _get_constructs(self, *default):
        '''Return the `Constructs` instance contained by the field construct.

.. versionadded:: 1.7.0

:Parameters:

    default: optional
        Return *default* if the `Constructs` instance has not been
        set.

:Returns:

    `Constucts`

**Examples:**

>>> c = f._get_contructs()

        '''
        return self._get_component('constructs', *default)
    #--- End: def

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def construct_type(self):
        '''Return a description of the construct type.
        
.. versionadded:: 1.7.0
        
:Returns:

    `str`
        The construct type.

        '''
        return 'field'
    #--- End: def        
        
    @property
    def domain(self):
        '''Return the domain.
        
``f.domain`` is equivalent to ``f.get_domain()`` 

.. versionadded:: 1.7.0

.. seealso:: `get_domain`

:Returns:

    `Domain`
        The domain.

**Examples:**

>>> d0 = f.domain
>>> d1 = f.get_domain()
>>> d0.equals(d1)
True
 
        '''
        return self.get_domain()
    #--- End: def
    
    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def del_data_axes(self, *default):
        '''Remove the keys of the domain axes spanned by the data array.

.. versionadded:: 1.7.0

.. seealso:: `data`, `get_data_axes`, `set_data_axes`

:Parameters:

    default: optional
        Return *default* if data axes have not been set.

:Returns:

    `tuple`
        The removed keys of the domain axes spanned by the data
        array. If unset then *default* is returned, if provided.

**Examples:**

>>> f.set_data_axes(['domainaxis0', 'domainaxis1'])
>>> f.get_data_axes()
('domainaxis0', 'domainaxis1')
>>> f.del_data_axes()
('domainaxis0', 'domainaxis1')
>>> print(f.del_dataxes(None))
None
>>> print(f.get_data_axes(None))
None

        '''
        return self._del_component('data_axes', *default)
    #--- End: def

    def get_domain(self):
        '''Return the domain.

.. versionadded:: 1.7.0

.. seealso:: `domain`

:Returns:

    `Domain`
        TODO

**Examples:**

>>> d = f.get_domain()

        '''
        return self._Domain.fromconstructs(self._get_constructs())
    #--- End: def

    def get_data_axes(self, *default):
        '''Return the identifiers of the domain axes spanned by the data
array.

.. versionadded:: 1.7.0

.. seealso:: `data`, `del_data_axes`, `get_data`, `set_data_axes`

:Parameters:

    default: optional
        Return *default* if data axes have not been set.

:Returns:

    `tuple` 
        The identifiers of the domain axes spanned by the data
        array. If unset then *default* is returned, if provided.

**Examples:**

>>> f.set_data_axes(['domainaxis0', 'domainaxis1'])
>>> f.get_data_axes()
('domainaxis0', 'domainaxis1')
>>> f.del_data_axes()
('domainaxis0', 'domainaxis1')
>>> print(f.del_dataxes(None))
None
>>> print(f.get_data_axes(None))
None

        '''    
        return self._get_component('data_axes', *default)
    #--- End: def
    
    def del_construct(self, key):
        '''Remove a metadata construct.

.. versionadded:: 1.7.0

.. seealso:: `get_construct`, `constructs`

:Parameters:

    key: `str`
        The construct identifier of the metadata construct to be
        removed.

        *Parameter example:*
          ``key='auxiliarycoordinate0'``
        
:Returns:

    metadata construct
        The removed metadata construct.

**Examples:**

>>> f.del_construct('auxiliarycoordinate0')

        '''
        if key in self.get_data_axes(()):
            raise ValueError(
                "Can't remove domain axis {!r} that is spanned by the field's data".format(key))

        return super().del_construct(key)
    #--- End: def

    def set_data(self, data, axes, copy=True):
        '''Set the data.

The units, calendar and fill value properties of the data object are
removed prior to insertion.

.. versionadded:: 1.7.0

.. seealso:: `data`, `del_data`, `get_data`, `has_data`,
             `set_data_axes`

:Parameters:

    data: `Data`
        The data to be inserted.

    axes: sequence of `str` or `None`, optional
        The identifiers of the domain axes spanned by the data
        array. If `None` then the data axes are not set.

        The axes may also be set afterwards with the `set_data_axes`
        method.

    copy: `bool`, optional
        If False then do not copy the data prior to insertion. By
        default the data are copied.

:Returns:

    `None`

**Examples:**

TODO

        '''
        if axes is not None:
            self.set_data_axes(axes)

        super().set_data(data, copy=copy)
    #--- End: def

    def set_data_axes(self, axes):
        '''Set the domain axes spanned by the data array.

.. versionadded:: 1.7.0

.. seealso:: `data`, `del_data_axes`, `get_data`, `get_data_axes`

:Parameters:

     axes: sequence of `str`
        The identifiers of the domain axes spanned by the data array.

        *Parameter example:*
          ``axes=['domainaxis0', 'domainaxis1']``

:Returns:

    `None`

**Examples:**

>>> f.set_data_axes(['domainaxis0', 'domainaxis1'])
>>> f.get_data_axes()
('domainaxis0', 'domainaxis1')
>>> f.del_data_axes()
('domainaxis0', 'domainaxis1')
>>> print(f.del_dataxes(None))
None
>>> print(f.get_data_axes(None))
None

        '''
        domain_axes = self.constructs(construct_type='domain_axis')
        for axis in axes:
            if axis not in domain_axes:
                raise ValueError(
"Can't set data axes: Domain axis {!r} doesn't exist".format(axis))
            
        self._set_component('data_axes', tuple(axes), copy=False)
    #--- End: def
    
#    def cell_methods(self, copy=False):
#        '''
#        '''
#        out = self.Constructs.cell_methods(copy=copy)
#
#        if not description:
#            return self.Constructs.cell_methods()
#        
#        if not isinstance(description, (list, tuple)):
#            description = (description,)
#            
#        cms = []
#        for d in description:
#            if isinstance(d, dict):
#                cms.append([self._CellMethod(**d)])
#            elif isinstance(d, basestring):
#                cms.append(self._CellMethod.parse(d))
#            elif isinstance(d, self._CellMethod):
#                cms.append([d])
#            else:
#                raise ValueError("asd 123948u m  BAD DESCRIPTION TYPE")
#        #--- End: for
#
#        keys = self.cell_methods().keys()                    
#        f_cell_methods = self.cell_methods().values()
#        nf = len(f_cell_methods)
#
#        out = {}
#        
#        for d in cms:
#            c = self._conform_cell_methods(d)
#
#            n = len(c)
#            for j in range(nf-n+1):
#                found_match = True
#                for i in range(0, n):
#                    if not f_cell_methods[j+i].match(c[i].properties()):
#                        found_match = False
#                        break
#                #--- End: for
#            
#                if not found_match:
#                    continue
#
#                # Still here?
#                key = tuple(keys[j:j+n])
#                if len(key) == 1:
#                    key = key[0]
#
#                if key not in out:
#                    value = f_cell_methods[j:j+n]
#                    if copy:
#                    value = [cm.copy() for cm in value]                        
#
#                out[key] = value
#            #--- End: for
#        #--- End: for
#        
#        return out
#    #--- End: def
    
#--- End: class



from future.utils import with_metaclass
from builtins import super

import abc

from copy import deepcopy

from . import Container


class Parameters(with_metaclass(abc.ABCMeta, Container)):
    '''Abstract base class for a collection of named parameters.

.. versionadded:: 1.7.0
    '''
    def __init__(self, parameters=None, source=None, copy=True): 
        '''**Initialization**

:Parameters:

    parameters: `dict`, optional
       Set parameters. The dictionary keys are parameter names, with
       corresponding values. Ignored if the *source* parameter is set.

       *Parameter example:*
         ``parameters={'earth_radius': 6371007.}``

       Parameters may also be set after initialisation with the
       `parameters` and `set_parameter` methods.

    source: optional
        Initialize the parameters from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__()

        self._set_component('parameters', {}, copy=False)

        if source:
            try:
                parameters = source.parameters()
            except AttributeError:
                parameters = None
        #--- End: if
        
        if parameters is None:
            parameters = {}
            copy = False

        self.parameters(parameters, copy=copy)
    #--- End: def

#    def __str__(self):
#        '''x.__str__() <==> str(x)
#
#        '''
#        out = []
#
#        parameters = self.parameters()
#        if parameters:
#            out.append('Parameters: {0}'.format(', '.join(sorted(parameters))))
#            
#        return '; '.join(out)
#    #--- End: def

    def copy(self):
        '''Return a deep copy.

``f.copy()`` is equivalent to ``copy.deepcopy(f)``.

.. versionadded:: 1.7.0

:Returns:

        The deep copy.

**Examples:**

>>> g = f.copy()

        '''
        return type(self)(source=self, copy=True)
    #--- End: def
    
    def del_parameter(self, parameter, *default):
        '''Delete a parameter.

.. versionadded:: 1.7.0

.. seealso:: `get_parameter`, `has_parameter`, `parameters`,
             `set_parameter`

:Parameters:

    parameter: `str`
        The name of the parameter to be deleted.

    default: optional
        Return *default* if the parameter has not been set.

:Returns:

        The removed parameter. If the parameter has not been then
        *default* is returned, if provided.

**Examples:**

        '''
        return self._get_component('parameters').pop(parameter, None)
    #--- End: def

    def get_parameter(self, term, *default):
        '''Get a parameter value.

.. versionadded:: 1.7.0

:Parameters:

    term: `str`
        The name of the term.

    default: optional

:Returns:

        The value of the term <SOMETING BAOUT DEFAULT> TODO

**Examples:**

>>> c.get_parameter('grid_north_pole_latitude')
70.0

>>> c.get_parameter('foo')
ERROR
>>> c.get_parameter('foo', 'nonexistent term')
'nonexistent term'


        '''
        d = self._get_component('parameters')
        if term in d:
            return d[term]
        
        if default:
            return default[0]

        raise AttributeError("{} doesn't have parameter term {!r}".format(
                self.__class__.__name__, term))
    #--- End: def
    
    def parameters(self, parameters=None, copy=True):
        '''Return or replace the parameter-valued terms.

.. versionadded:: 1.7.0

.. seealso:: `ancillaries`

:Parameters:

    parameters: `dict`, optional
        Replace all parameter-valued terms with those provided.

        *Parameter example:*
          ``parameters={'earth_radius': 6371007}``

    copy: `bool`, optional
        TODO

:Returns:

    `dict`
        The parameter-valued terms and their values. If the
        *parameters* keyword has been set then the parameter-valued
        terms prior to replacement are returned.

**Examples:**

>>> c.parameters()
{'standard_parallel': 25.0;
 'longitude_of_central_meridian': 265.0,
 'latitude_of_projection_origin': 25.0}

>>> c.parameters()
{}

        '''
        out = self._get_component('parameters').copy()

        if parameters is not None:
            if copy:
                properties = deepcopy(parameters)
            else:
                properties = parameters.copy()
                
            self._set_component('parameters', parameters, copy=False)

        return out
    #--- End: def

    def set_parameter(self, term, value, copy=True):
        '''Set a parameter-valued term.

.. versionadded:: 1.7.0

.. seealso:: `parameters`

:Returns:

    `None`

**Examples:**

>>> c.parameters()
{'standard_parallel': 25.0;
 'latitude_of_projection_origin': 25.0}
>>> c.set_parameter('longitude_of_central_meridian', 265.0)
>>> c.parameters()
{'standard_parallel': 25.0;
 'longitude_of_central_meridian': 265.0,
 'latitude_of_projection_origin': 25.0}
        '''
        if copy:
            value = deepcopy(value)
            
        self._get_component('parameters')[term] = value
    #--- End: def
    
#--- End: class



from builtins import object


class RaggedContiguous(object):
    '''TODO

.. versionadded:: 1.7.0
    '''

    def get_count_variable(self, *default):
        '''TODO

.. versionadded:: 1.7.0

:Parameters:

    default: optional
        Return *default* if the count variable has not been set.

:Returns:

        TODO

**Examples:**

TODO
        '''
        return self._get_component('count_variable', *default)
#        try:
#            return self._count_variable
#        except AttributeError:
#            if default:
#                return default[0]
#
#            raise AttributeError("{!r} has no count variable".format(
#                self.__class__.__name__))
    #--- End: def

#--- End: class


class RaggedIndexed(object):
    '''TODO

    '''
    def get_index_variable(self, *default):
        '''TODO

.. versionadded:: 1.7.0

:Parameters:

    default: optional
        Return *default* if the index variable has not been set.

:Returns:

        TODO

**Examples:**

TODO

        '''
        return self._get_component('index_variable', *default)
#        try:
#            return self._index_variable
#        except AttributeError:
#            if default:
#                return default[0]
#
#            raise AttributeError("{!r} has no index variable".format(
#                self.__class__.__name__))
    #--- End: def
    
#--- End: class



from builtins import (str, super) #(object, str)
from future.utils import with_metaclass

import abc

from ...abstract import Container


class Array(with_metaclass(abc.ABCMeta, Container)):
    '''Abstract base class for a container of an array.

The form of the array is defined by the initialization parameters of a
subclass.

See `cfdm.core.NumpyArray` for an example implementation.

.. versionadded:: 1.7.0

    '''
    def __init__(self, **kwargs):
        '''**Initialization**

:Parameters:

    kwargs: *optional*
        Named parameters and their values that define the array.

        '''
        super().__init__()

        for key, value in kwargs.items():
            self._set_component(key, value, copy=False)
    #--- End: def

    def __array__(self, *dtype):
        '''The numpy array interface.

.. versionadded:: 1.7.0

:Returns: 

    `numpy.ndarray`
        An independent numpy array of the data.

**Examples:**

>>> isinstance(a, Array)
True
>>> n = numpy.asanyarray(a)
>>> isinstance(n, numpy.ndarray)
True

        '''
        array = self.get_array()
        if not dtype:
            return array
        else:
            return array.astype(dtype[0], copy=False)
    #--- End: def

    def __deepcopy__(self, memo):
        '''Called by the `copy.deepcopy` function.

x.__deepcopy__() <==> copy.deepcopy(x)

Copy-on-write is employed. Therefore, after copying, care must be
taken when making in-place modifications to attributes of either the
original or the new copy.

.. versionadded:: 1.7.0

**Examples:**

>>> import copy
>>> y = copy.deepcopy(x)

        '''
        return self.copy()
    #--- End: def

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    @abc.abstractmethod
    def dtype(self):
        '''Data-type of the data elements.

.. versionadded:: 1.7.0

**Examples:**

>>> a.dtype
dtype('float64')
>>> print(type(a.dtype))
<type 'numpy.dtype'>

        '''
        raise NotImplementedError()
    #--- End: def

    @property
    @abc.abstractmethod
    def ndim(self):
        '''Number of array dimensions

.. versionadded:: 1.7.0

**Examples:**

>>> a.shape
(73, 96)
>>> a.ndim
2
>>> a.size
7008

>>> a.shape
(1, 1, 1)
>>> a.ndim
3
>>> a.size
1

>>> a.shape
()
>>> a.ndim
0
>>> a.size
1

        '''
        raise NotImplementedError()
    #--- End: def
    
    @property
    @abc.abstractmethod
    def shape(self):
        '''Tuple of array dimension sizes.

.. versionadded:: 1.7.0

**Examples:**

>>> a.shape
(73, 96)
>>> a.ndim
2
>>> a.size
7008

>>> a.shape
(1, 1, 1)
>>> a.ndim
3
>>> a.size
1

>>> a.shape
()
>>> a.ndim
0
>>> a.size
1

        '''
        raise NotImplementedError()
    #--- End: def
    
    @property
    @abc.abstractmethod
    def size(self):
        '''Number of elements in the array.

.. versionadded:: 1.7.0

**Examples:**

>>> a.shape
(73, 96)
>>> a.size
7008
>>> a.ndim
2

>>> a.shape
(1, 1, 1)
>>> a.ndim
3
>>> a.size
1

>>> a.shape
()
>>> a.ndim
0
>>> a.size
1

        '''
        raise NotImplementedError()
    #--- End: def

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self):
        '''Return a deep copy of the array.

``a.copy() is equivalent to ``copy.deepcopy(a)``.

Copy-on-write is employed. Therefore, after copying, care must be
taken when making in-place modifications to attributes of either the
original or the new copy.

.. versionadded:: 1.7.0

:Returns:

        The deep copy.

**Examples:**

>>> b = a.copy()

        '''
        klass = self.__class__
        new = klass.__new__(klass)
        new.__dict__ = self.__dict__.copy()
        return new
    #--- End: def

    @abc.abstractmethod
    def get_array(self):
        '''Return an independent numpy array containing the data.

.. versionadded:: 1.7.0

:Returns:

    `numpy.ndarray`
        An independent numpy array of the data.

**Examples:**

>>> n = a.get_array()
>>> isinstance(n, numpy.ndarray)
True

        '''
        raise NotImplementedError()
    #--- End: def
    
#--- End: class



from future.utils import with_metaclass
from builtins import str

import abc

import numpy

from ...core.data import Array as core_Array


class Array(with_metaclass(abc.ABCMeta, core_Array)):
    '''Abstract base class for a container of an underlying array.

The form of the array is defined by the initialization parameters of a
subclass.

See `cfdm.NumpyArray` for an example implementation.

.. versionadded:: 1.7.0

    '''
    @abc.abstractmethod
    def __getitem__(self, indices):
        '''Return a subspace as an independent numpy array.

x.__getitem__(indices) <==> x[indices]

Indexing follows rules that are very similar to the numpy indexing
rules, the only differences being:

* An integer index i takes the i-th element but does not reduce the
  rank by one.

..

* When two or more dimensions' indices are sequences of integers then
  these indices work independently along each dimension (similar to
  the way vector subscripts work in Fortran). This is the same
  behaviour as indexing on a Variable object of the netCDF4 package.

.. versionadded:: 1.7.0

        '''
        raise NotImplementedError()
    #--- End: def

    def __repr__(self):
        '''Called by the `repr` built-in function.

x.__repr__() <==> repr(x)

.. versionadded:: 1.7.0

        '''      
        return "<{0}: {1}>".format(self.__class__.__name__, str(self))
    #--- End: def
        
    def __str__(self):
        '''Called by the `str` built-in function.

x.__str__() <==> str(x)

.. versionadded:: 1.7.0

        '''
        return "shape={0}, dtype={1}".format(self.shape, self.dtype)
    #--- End: def

    def get_compression_type(self):
        '''The type of compression that has been applied to the underlying
array.

.. versionadded:: 1.7.0

:Returns:

    `str`
        The compression type. An empty string means that no
        compression has been applied.

**Examples:**

>>> a.compression_type
''

>>> a.compression_type
'gathered'

>>> a.compression_type
'ragged contiguous'

        '''
        return self._get_component('compression_type', '')
    #--- End: def

    def get_compressed_array(self, *default):
        '''Return an independent numpy array containing of the underlying
compressed array.

.. versionadded:: 1.7.0

:Returns:

    `numpy.ndarray`
        The compressed array.

**Examples:**

>>> n = a.get_compressed_array()
>>> isinstance(n, numpy.ndarray)
True

        '''
        return self._get_component('compressed_array').get_compressed_array()
    #--- End: def
    
    def get_compressed_axes(self):
        '''Return axes that are compressed in the underlying array.

.. versionadded:: 1.7.0

:Returns:

    `list`
        The compressed axes described by their integer positions in
        the list of uncompressed axes. If no axes have been compressed
        then the list is empty.

**Examples:**

        '''
        return []
    #--- End: def

    def get_compressed_dimension(self, *default):
        '''Return the position of the compressed dimension in the compressed
array.

.. versionadded:: 1.7.0

.. seealso:: `get_compressed_axearray`, `get_compressed_axes`,
             `get_compressed_type`

:Parameters:

    default: optional
        Return *default* if the underlying array is not compressed.

:Returns:

    `int`
        The position of the compressed dimension in the compressed
        array. If the underlying is not compressed then *default* is
        returned, if provided.

**Examples:**

>>> i = d.get_compressed_dimension()

        '''
        return self._get_component('compressed_dimension', *default)
    #--- End: def

    @classmethod
    def get_subspace(cls, array, indices, copy=True):
        '''Return a subspace, defined by indices, of a numpy array.

Only certain type of indices are allowed. See the *indices* parameter
for details.

Indexing is similar to numpy indexing. Given the restrictions on the
type of indices allowed - see the *indicies* parameter - the only
difference to numpy indexing is

  * When two or more dimension's indices are sequences of integers
    then these indices work independently along each dimension
    (similar to the way vector subscripts work in Fortran).

.. versionadded:: 1.7.0

:Parameters:

    array: `numpy.ndarray`
        The array to be subspaced.
        
    indices: 
        The indices that define the subspace.
     ..
        Must be either `Ellipsis` or a sequence that contains an index
        for each dimension. In the latter case, each dimension's index
        must either be a `slice` object or a sequence of two or more
        integers.

          *Parameter example:*
            indices=Ellipsis
  
          *Parameter example:*
            indices=[[5, 7, 8]]
  
          *Parameter example:*
            indices=[slice(4, 7)]

          *Parameter example:*
            indices=[slice(None), [5, 7, 8]]
  
          *Parameter example:*
            indices=[[2, 5, 6], slice(15, 4, -2), [8, 7, 5]]

    copy: `bool`
        If `False` then the returned subspace may (or may not) be
        independent of the input *array*. By default the returned
        subspace is independent of the input *array*.

:Returns:

    `numpy.ndarray`

        '''
        if indices is not Ellipsis:
#        if indices is Ellipsis:
#            pass
#        elif not isinstance(indices, tuple):
#            array = array[indices]
#        else:
            axes_with_list_indices = [i for i, x in enumerate(indices)
                                      if not isinstance(x, slice)]
            n_axes_with_list_indices = len(axes_with_list_indices)
        
            if n_axes_with_list_indices < 2:
                # ----------------------------------------------------
                # At most one axis has a list-of-integers index so we
                # can do a normal numpy subspace
                # ----------------------------------------------------
                array = array[tuple(indices)]
            else:
                # ----------------------------------------------------
                # At least two axes have list-of-integers indices so
                # we can't do a normal numpy subspace
                # ----------------------------------------------------
                if numpy.ma.isMA(array):
                    take = numpy.ma.take
                else:
                    take = numpy.take
        
                indices = list(indices)
                for axis in axes_with_list_indices:
                    array = take(array, indices[axis], axis=axis)
                    indices[axis] = slice(None)
        
                if n_axes_with_list_indices < len(indices):
                    # Apply subspace defined by slices
                    array = array[tuple(indices)]
        #--- End: if

        if copy:
            if numpy.ma.isMA(array) and not array.ndim:
                # This is because numpy.ma.copy doesn't work for
                # scalar arrays (at the moment, at least)
                ma_array = numpy.ma.empty((), dtype=array.dtype)
                ma_array[...] = array
                array = ma_array
            else:
                array = array.copy()
        #--- End: if
        
        return array                
    #--- End: def

#--- End: class



from future.utils import with_metaclass

import abc

from . import PropertiesDataBounds


class Coordinate(with_metaclass(abc.ABCMeta, PropertiesDataBounds)):
    '''Abstract base class for dimension and auxiliary coordinate
constructs of the CF data model.

.. versionadded:: 1.7.0

    '''
#--- End: class



from builtins import (str, super)
from past.builtins import basestring

from copy import deepcopy

from . import abstract


class CellMethod(abstract.Properties):
    '''A cell method construct of the CF data model.

One or more cell method constructs describe how the cell values of the
field construct represent the variation of the physical quantity
within its cells, i.e. the structure of the data at a higher
resolution.

A single cell method construct consists of a set of axes, a "method"
property which describes how a value of the field construct's data
array describes the variation of the quantity within a cell over those
axes (e.g. a value might represent the cell area average), and
properties serving to indicate more precisely how the method was
applied (e.g. recording the spacing of the original data, or the fact
that the method was applied only over El Nino years).

.. versionadded:: 1.7.0

    '''
    def __init__(self, axes=None, method=None, properties=None,
                 source=None, copy=True):
        '''**Initialisation**

:Parameters:

    axes: (sequence of) `str`, optional
        Set the axes of the cell method construct, specified either by
        the construct identifiers of domain axis constructs, standard
        names, or the special string ``'area'``.

          *Parameter example:*
             ``axes='domainaxis0'``
        
          *Parameter example:*
             ``axes=['domainaxis0']``
        
          *Parameter example:*
             ``axes=('domainaxis0', 'domainaxis1')``
        
          *Parameter example:*
             ``axes='area'``
        
          *Parameter example:*
             ``axes=['domainaxis2', 'time']``
        
         The axes may also be set after initialisation with the
         `set_axes` method.

    method: `str`, optional
        Set the axes of the cell method construct. Either one or more
        domain axis construct identifiers or standard names. Ignored
        if the *source* parameter is set.

          *Parameter example:*
             ``method='mean'

         The method may also be set after initialisation with the
         `set_method` method.

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
          ``properties={'comment': 'sampled instantaneously'}``
       
        *Parameter example:*
          ``properties={'where': 'sea', ''over': 'ice'}``
        
        Properties may also be set after initialisation with the
        `properties` and `set_property` methods.

    source: optional
        Initialize the axes, method and properties from those of
        *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization By default parameters are deep copied.

        '''
        super().__init__(properties=properties, source=source,
                         copy=copy)

        if source:
            try:
                axes = source.get_axes(None)
            except AttributeErrror:
                axes = None              

            try:
                method = source.get_method(None)
            except AttributeErrror:
                method = None              
        #--- End: if

        if axes is not None:                
            axes = self.set_axes(axes)

        if method is not None:                
            method = self.set_method(method)
    #--- End: def

    @property
    def construct_type(self):
        '''Return a description of the construct type.
        
.. versionadded:: 1.7.0
        
:Returns:

    `str`
        The construct type.

        '''
        return 'cell_method'
    #--- End: def

    def del_axes(self, *default):
        '''Remove the axes of the cell method.

.. versionadded:: 1.7.0

.. seealso:: `get_axes`, `has_axes`, `set_axes`

:Parameters:

    default: optional
        Return *default* if axes have not been set.

:Returns:

    `tuple` or *default*
        The removed axes, identified by field key or standard name. If
        unset then *default* is returned, if provided.

**Examples:**

>>> c.set_axes('domainaxis1')
>>> c.has_axes()
True
>>> c.get_axes()
('domainaxis1',)
>>> c.del_axes()
>>> c.has_axes()
False
>>> c.get_axes('NO AXES')
'NO AXES'
>>> c.del_axes('NO AXES')
'NO AXES'

        '''
        return self._del_component('axes')
    #--- End: def
    
    def del_method(self, *default):
        '''Remove the method of the cell method.

.. versionadded:: 1.7.0

.. seealso:: `get_method`, `has_method`, `set_method`

:Parameters:

    default: optional
        Return *default* if method has not been set.

:Returns:

    `tuple` or *default*
        The removed method. If unset then *default* is returned, if
        provided.

**Examples:**

>>> c.set_method('minimum')
>>> c.has_method()
True
>>> c.get_method()
('time',)
>>> c.del_method()
>>> c.has_method()
False
>>> c.get_method('NO METHOD')
'NO METHOD'
>>> c.del_method('NO METHOD')
'NO METHOD'

        '''
        return self._del_component('method', *default)
    #--- End: def
    
    def get_axes(self, *default):
        '''Return the axes of the cell method.

.. versionadded:: 1.7.0

.. seealso:: `del_axes`, `has_axes`, `set_axes`

:Parameters:

    default: optional
        Return *default* if axes have not been set.

:Returns:

    `tuple`
        The axes. If axes have not been set then return the value
        of *default* parameter, if provided.

**Examples:**

>>> c.set_method('minimum')
>>> c.has_method()
True
>>> c.get_method()
'minimum'
>>> c.del_method()
>>> c.has_method()
False
>>> c.get_method('NO METHOD')
'NO METHOD'
>>> c.del_method('NO METHOD')
'NO METHOD'

        '''
        return self._get_component('axes', *default)
    #--- End: def

    def get_method(self, *default):
        '''Return the method of the cell method.

.. versionadded:: 1.7.0

.. seealso:: `del_method`, `has_method`, `set_method`

:Parameters:

    default: optional
        Return *default* if the "method" propoerty has not been set.

:Returns:

    `tuple`
        The method. If the method has not been set then return the
        value of *default* parameter, if provided.

**Examples:**

>>> c.set_method('minimum')
>>> c.has_method()
True
>>> c.get_method()
'minimum'
>>> c.del_method()
>>> c.has_method()
False
>>> c.get_method('NO METHOD')
'NO METHOD'
>>> c.del_method('NO METHOD')
'NO METHOD'

        '''
        return self._get_component('method', *default)
    #--- End: def

    def has_axes(self):
        '''Whether the axes of the cell method have been set.

.. versionadded:: 1.7.0

.. seealso:: `del_axes`, `get_axes`, `set_axes`

:Returns:

     `bool`
        True if the axes have been set, otherwise False.

**Examples:**

>>> c.set_axes('domainaxis1')
>>> c.has_axes()
True
>>> c.get_axes()
('domainaxis1',)
>>> c.del_axes()
>>> c.has_axes()
False
>>> c.get_axes('NO AXES')
'NO AXES'
>>> c.del_axes('NO AXES')
'NO AXES'

'''
        return self._has_component('axes')
    #--- End: def

    def has_method(self):
        '''Whether the method of the cell method has been set.

.. versionadded:: 1.7.0

.. seealso:: `del_axxes`, `get_method`, `set_method`

:Returns:

     `bool`
        True if the method has been set, otherwise False.

**Examples:**

>>> c.set_method('minimum')
>>> c.has_method()
True
>>> c.get_method()
'minimum'
>>> c.del_method()
>>> c.has_method()
False
>>> c.get_method('NO METHOD')
'NO METHOD'
>>> c.del_method('NO METHOD')
'NO METHOD'

'''
        return self._has_component('method')
    #--- End: def

    def set_axes(self, value, copy=True):
        '''Set the axes of the cell method.

.. versionadded:: 1.7.0

.. seealso:: `del_axes`, `get_axes`, `has_axes`

:Parameters:

    value: (sequence of) `str`
        The axes, specified either by the construct identifiers of
        domain axis constructs, standard names, or the special string
        ``'area'``.

        *Parameter example:*
          ``axes='domainaxis0'``

        *Parameter example:*
          ``axes='time'``

        *Parameter example:*
          ``axes='area'``

        *Parameter example:*
          ``axes=['domainaxis0', 'domainaxis2']``

        *Parameter example:*
          ``axes=['time', 'area']``

        *Parameter example:*
          ``axes=['domainaxis0', 'time']``

:Returns:

     `None`

**Examples:**

>>> c.set_axes('domainaxis1')
>>> c.has_axes()
True
>>> c.get_axes()
('domainaxis1',)
>>> c.del_axes()
>>> c.has_axes()
False
>>> c.get_axes('NO AXES')
'NO AXES'
>>> c.del_axes('NO AXES')
'NO AXES'

>>> c.set_axes(['domainaxis1', 'domainaxis0'])

>>> c.set_axes(['time', 'domainaxis0'])

>>> c.set_axes('time')

        '''
        if copy:
            value = deepcopy(value)
        
        if isinstance(value, basestring):
            value = (value,)
        else:
            value = tuple(value)
            
        return self._set_component('axes', value, copy=False)
    #--- End: def

    def set_method(self, value, copy=True):
        '''Set the method of the cell method.

.. versionadded:: 1.7.0

.. seealso:: `del_method`, `get_method`, `has_method`

:Parameters:

    value: `str`
        The method.

        *Parameter example:*
          ``method='variance'``

:Returns:

     `None`

**Examples:**

>>> c.set_method('minimum')
>>> c.has_method()
True
>>> c.get_method()
'minimum'
>>> c.del_method()
>>> c.has_method()
False
>>> c.get_method('NO METHOD')
'NO METHOD'
>>> c.del_method('NO METHOD')
'NO METHOD'

        '''
        return self._set_component('method', value, copy=copy)
    #--- End: def

#--- End: class



from builtins import super
from future.utils import with_metaclass

import abc

from . import PropertiesData


class PropertiesDataBounds(with_metaclass(abc.ABCMeta, PropertiesData)):
    '''Abstract base class for a data array with bounds and descriptive
properties.

.. versionadded:: 1.7.0

    '''    
    def __init__(self, properties=None, data=None, bounds=None,
                 geometry=None, interior_ring=None, source=None,
                 copy=True, _use_data=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
           ``properties={'standard_name': 'longitude'}``
        
        Properties may also be set after initialisation with the
        `properties` and `set_property` methods.
  
    data: `Data`, optional
        Set the data array. Ignored if the *source* parameter is set.
        
        The data array may also be set after initialisation with the
        `set_data` method.
  
    bounds: `Bounds`, optional
        Set the bounds array. Ignored if the *source* parameter is
        set.
        
        The bounds array may also be set after initialisation with the
        `set_bounds` method.
  
    geometry: `str`, optional
        Set the geometry type. Ignored if the *source* parameter is
        set.
        
        *Parameter example:*
           ``geometry='polygon'``
        
        The geometry type may also be set after initialisation with
        the `set_geometry` method.
  
    interior_ring: `InteriorRing`, optional
        Set the interior ring array. Ignored if the *source* parameter
        is set.
        
        The interior ring array may also be set after initialisation
        with the `set_interior_ring` method.
  
    source: optional
        Initialize the properties, data and bounds from those of
        *source*.
 
    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        # Initialise properties and data
        super().__init__(properties=properties, data=data,
                         source=source, copy=copy,
                         _use_data=_use_data)

        # Get bounds, geometry type and interior ring from source
        if source is not None:
            try:
                bounds = source.get_bounds(None)
            except AttributeError:
                bounds = None
                
            try:
                geometry = source.get_geometry(None)
            except AttributeError:
                geometry = None
                
            try:
                interior_ring = source.get_interior_ring(None)
            except AttributeError:
                interior_ring = None
        #--- End: if

        # Initialise bounds
        if bounds is not None:
            if copy or not _use_data:
                bounds = bounds.copy(data=_use_data)
                
            self.set_bounds(bounds, copy=False)

        # Initialise interior ring
        if interior_ring is not None:
            if copy or not _use_data:
                interior_ring = interior_ring.copy(data=_use_data)
                
            self.set_interior_ring(interior_ring, copy=False)
            
        # Initialise the geometry type
        if geometry is not None:
            self.set_geometry(geometry)
    #--- End: def

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def bounds(self):
        '''Return the bounds.

``f.bounds`` is equivalent to ``f.get_bounds()`` 

Note that a `Bounds` instance is returned. Use the `get_array` method
of the `Bounds` instance to return the bounds as a `numpy` array.

.. versionadded:: 1.7.0

.. seealso:: `data`, `del_bounds`, `get_bounds`, `has_bounds`,
             `set_bounds`

:Returns:

    `Bounds`
        The bounds.

**Examples:**

>>> import numpy
>>> b = cfdm.Bounds(data=cfdm.Data(numpy.arange(10).reshape(5, 2)))
>>> c.set_bounds(b)
>>> c.has_bounds()
True
>>> b = c.bounds
>>> b
<Bounds: (5, 2) >
>>> b.data
<Data(5, 2): [[0, ..., 9]]>
>>> b.data.shape
(5, 2)
        '''
        return self.get_bounds()
    #--- End: def

    @property
    def interior_ring(self):
        '''TODO

.. versionadded:: 1.8
        '''
        return self.get_interior_ring()
    #--- End: def

#    # ----------------------------------------------------------------
#    # Methods
#    # ----------------------------------------------------------------
#    def copy(self, data=True):
#        '''Return a deep copy.
#
#``c.copy()`` is equivalent to ``copy.deepcopy(c)``.
#
#`Data` instances contained within are copied with a copy-on-write
#technique. This means that a copy takes up very little extra memory,
#even when the original contains very large data arrays, and the copy
#operation is fast.
#
#.. versionadded:: 1.7.0
#
#:Parameters:
#
#    data: `bool`, optional
#        If False then do not copy the data, bounds nor interior ring
#        arrays. By default they are copied.
#
#:Returns:
#
#        The deep copy.
#
#**Examples:**
#
#>>> g = f.copy()
#>>> g = f.copy(data=False)
#>>> g.has_data()
#False
#>>> g.bounds.has_data()
#False
#
#        '''
#        return super().copy(data=data)
#    #--- End: def

    def del_bounds(self, *default):
        '''Remove the bounds.

.. versionadded:: 1.7.0

.. seealso:: `del_data`, `get_bounds`, `has_bounds`, `set_bounds`

:Parameters:

    default: optional
        Return *default* if the bounds has not been set.

:Returns:

        The removed bounds. If the bounds has not been set then the
        *default* is returned, if provided.

**Examples:**

>>> import numpy
>>> b = cfdm.Bounds(data=cfdm.Data(numpy.arange(10).reshape(5, 2)))
>>> c.set_bounds(b)
>>> c.has_bounds()
True
>>> c.get_bounds()
<Bounds: (5, 2) >
>>> b = c.del_bounds()
>>> b
<Bounds: (5, 2) >
>>> c.has_bounds()
False
>>> print(c.get_bounds(None))
None
>>> print(c.del_bounds(None))
None

        '''
        return self._del_component('bounds', *default)
    #--- End: def

    def del_geometry(self, *default):
        '''Delete the geometry type. TODO

.. versionadd:: 1.8

.. seealso:: `get_geometry`, `has_geometry`, `set_geometry`

:Parameters:

    default: optional
        Return *default* if the geometry type has not been set.

:Returns:

        The removed geometry type. If the geometry type has not been
        set then the *default* is returned, if provided.

**Examples:**

        '''
        return self._del_component('geometry', *default)
    #--- End: def

    def get_bounds(self, *default):
        '''Return the bounds.

.. versionadd:: 1.7.0

.. seealso:: `bounds`, `get_data`, `del_bounds`, `has_bounds`,
             `set_bounds`

:Parameters:

    default: optional
        Return *default* if and only if the bounds have not been set.

:Returns:

        The bounds. If the bounds have not been set, then return the
        value of *default* parameter if provided.

**Examples:**

>>> import numpy
>>> b = cfdm.Bounds(data=cfdm.Data(numpy.arange(10).reshape(5, 2)))
>>> c.set_bounds(b)
>>> c.has_bounds()
True
>>> c.get_bounds()
<Bounds: (5, 2) >
>>> b = c.del_bounds()
>>> b
<Bounds: (5, 2) >
>>> c.has_bounds()
False
>>> print(c.get_bounds(None))
None
>>> print(c.del_bounds(None))
None

        '''
        return self._get_component('bounds', *default)
    #--- End: def

    def get_geometry(self, *default):
        '''Return the geometry type.

.. versionadded:: 1.8

.. seealso:: `get_array`, `get_data`, `has_bounds`, `set_bounds`

:Parameters:

    default: optional
        Return *default* if and only if the bounds have not been set.

:Returns:

        TODO

**Examples:**

TODO
        '''
        return self._get_component('geometry', *default)
    #--- End: def

    def get_interior_ring(self, *default):
        '''Return the interior_ring.

.. versionadded:: 1.8

.. seealso:: `del_interior_ring`, `get_bounds`, `has_interior_ring`,
             `set_interior_ring`

:Parameters:

    default: optional
        Return *default* if and only if the HHHh have not been set.

:Returns:

        TODO

**Examples:**

>>> i = c.get_interior_ring()
        '''
        return self._get_component('interior_ring', *default)
    #--- End: def

    def has_bounds(self):
        '''Whether or not there are bounds.
        
.. versionadded:: 1.7.0

.. seealso:: `del_bounds`, `get_bounds`, `has_data`, `set_bounds`

:Returns:

    `bool`
        True if there are bounds, otherwise False.

**Examples:**

>>> import numpy
>>> b = cfdm.Bounds(data=cfdm.Data(numpy.arange(10).reshape(5, 2)))
>>> c.set_bounds(b)
>>> c.has_bounds()
True
>>> c.get_bounds()
<Bounds: (5, 2) >
>>> b = c.del_bounds()
>>> b
<Bounds: (5, 2) >
>>> c.has_bounds()
False
>>> print(c.get_bounds(None))
None
>>> print(c.del_bounds(None))
None

        '''
        return self._has_component('bounds')
    #--- End: def

    def has_geometry(self):
        '''True if there is a goemetry type. TODO
        
.. versionadded:: 1.8

.. seealso:: `del_bounds`, `get_bounds`, `has_data`, `set_bounds`

:Returns:

    `bool`

**Examples:**

>>> x = c.has_geometry()

        '''
        return self._has_component('geometry')
    #--- End: def

    def has_interior_ring(self):
        '''True if there are interior_ring. TODO
        
.. versionadded:: 1.8

.. seealso:: `del_interior_ring`, `get_interior_ring`, `has_data`, `set_interior_ring`

:Returns:

    `bool`
        True if there are interior_ring, otherwise False.

**Examples:**

>>> x = f.has_interior_ring()
>>> if c.has_interior_ring():
...     print 'Has interior_ring'

        '''
        return self._has_component('interior_ring')
    #--- End: def

    def set_bounds(self, bounds, copy=True):
        '''Set the bounds.

.. versionadded:: 1.7.0

.. seealso: `del_bounds`, `get_bounds`, `has_bounds`, `set_data`

:Parameters:

    data: `Bounds`
        The bounds to be inserted.

    copy: `bool`, optional
        If False then do not copy the bounds prior to insertion. By
        default the bounds are copied.

:Returns:

    `None`

**Examples:**

>>> import numpy
>>> b = cfdm.Bounds(data=cfdm.Data(numpy.arange(10).reshape(5, 2)))
>>> c.set_bounds(b)
>>> c.has_bounds()
True
>>> c.get_bounds()
<Bounds: (5, 2) >
>>> b = c.del_bounds()
>>> b
<Bounds: (5, 2) >
>>> c.has_bounds()
False
>>> print(c.get_bounds(None))
None
>>> print(c.del_bounds(None))
None

        '''
        if copy:
            bounds = bounds.copy()

        self._set_component('bounds', bounds, copy=False)
    #--- End: def

    def set_geometry(self, value, copy=True):
        '''Set the geometry type.

.. versionadded:: 1.8

.. seealso: `del_bounds`, `get_bounds`, `has_bounds`, `set_data`

:Parameters:

    value: `str`

:Returns:

    `None`

**Examples:**
        '''
        self._set_component('geometry', value, copy=copy)
    #--- End: def

    def set_interior_ring(self, interior_ring, copy=True):
        '''Set the interior_ring.

.. versionadded:: 1.8

.. seealso: `del_interior_ring`, `get_interior_ring`,
            `has_interior_ring`

:Parameters:

    interior_ring: `InteriorRing`
        The interior_ring to be inserted.

    copy: `bool`, optional
        If False then do not copy the interior_ring prior to insertion. By
        default the interior_ring are copied.

:Returns:

    `None`

**Examples:**

>>> c.set_interior_ring(b)
        '''
        if copy:
            interior_ring = interior_ring.copy()

        self._set_component('interior_ring', interior_ring, copy=False)
    #--- End: def

#--- End: class



from builtins import (object, super)

from . import mixin
from . import core


class List(mixin.NetCDFVariable,
           mixin.PropertiesData,
           core.abstract.PropertiesData):
    '''A list variable required to uncompress a gathered array.

Compression by gathering combines axes of a multidimensional array
into a new, discrete axis whilst omitting the missing values and thus
reducing the number of values that need to be stored.

The information needed to uncompress the data is stored in a list
variable that gives the indices of the required points.

.. versionadded:: 1.7.0

    '''
    def __init__(self, properties=None, data=None, source=None,
                 copy=True, _use_data=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
          ``properties={'long_name': 'uncompression indices'}``

        Properties may also be set after initialisation with the
        `properties` and `set_property` methods.

    data: `Data`, optional
        Set the data array. Ignored if the *source* parameter is set.

        The data array may also be set after initialisation with the
        `set_data` method.

    source: optional
        Initialize the properties and data from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(properties=properties, data=data,
                         source=source, copy=copy,
                         _use_data=_use_data)
        
        self._initialise_netcdf(source)
    #--- End: def
    
    def dump(self, display=True, field=None, key=None, _title=None,
             _create_title=True, _prefix='', _level=0,
             _omit_properties=None):
        '''A full description of the list variable.

Returns a description of all properties, including those of
components, and provides selected values of all data arrays.

.. versionadded:: 1.7.0

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.

:Returns:

    `None` or `str`
        The description. If *display* is True then the description is
        printed and `None` is returned. Otherwise the description is
        returned as a string.

        '''
        if _create_title and _title is None: 
            _title = 'List: ' + self.name(default='')

        return super().dump(display=display, field=field, key=key,
                            _omit_properties=_omit_properties,
                            _prefix=_prefix, _level=_level,
                            _title=_title,
                            _create_title=_create_title)
    #--- End: def
    
#--- End: class
